#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['TestCase', 'TestResult', 'TestSuite', 'TextTestRunner', 'main',
           'nottest', 'category']

from inspect import getfile
import stat
import os
import sys
import filecmp
import re
if sys.version_info[:2] < (2, 7):
    try:
        import unittest2 as unittest
        main = unittest.main
        using_unittest2 = True
    except ImportError:
        import unittest
        main = unittest.main
        using_unittest2 = False
else:
    import unittest
    using_unittest2 = True
    main = unittest.main

from six import iteritems, itervalues

#
# Defer the pyutilib.misc import until it is actually needed.  If we
# import it here, then PyUtilib's test coverage will report all
# "module-level" lines as uncovered because they were executed before
# the nose coverage plugin was watching things.
#
#import pyutilib.misc

if using_unittest2:
    __all__.extend(
        ['skip', 'skipIf', 'skipUnless', 'expectedFailure', 'SkipTest'])
    skip = unittest.skip
    skipIf = unittest.skipIf
    skipUnless = unittest.skipUnless
    expectedFailure = unittest.expectedFailure
    SkipTest = unittest.SkipTest
import subprocess

TextTestRunner = unittest.TextTestRunner
TestResult = unittest.TestResult
TestSuite = unittest.TestSuite

try:
    from nose.tools import nottest
except ImportError:

    def nottest(func):
        """Decorator to mark a function or method as *not* a test"""
        func.__test__ = False
        return func

def _category_to_tuple(_cat):
    _cat = str(_cat).lower().strip()
    if _cat.endswith('=0') or _cat.endswith('=1'):
        _val = int(_cat[-1])
        _cat = _cat[:-2]
    else:
        _val = 1
    if _cat and _cat[0] in '!~-':
        _val = 1 - _val
        _cat = _cat[1:]
    return _cat, _val

def category(*args, **kwargs):
    # Get the set of categories for this test
    _categories = {}
    for cat in args:
        _cat, _val = _category_to_tuple(cat)
        if not _cat:
            continue
        _categories[_cat] = _val

    # Note: we used to try and short-circuit the nosetest test selection
    # and return test skips for tests that couldn't/wouldn't be run.
    # However, this code was unreliable, as categories could be set by
    # both decorating the TestCase (class) and the function.  As a
    # result, we will just rely on nosetest to do the right thing.

    def _id(func):
        if hasattr(func, '__mro__') and TestCase in func.__mro__:
            # @category() called on a TestCase class
            if len(_categories) > (1 if 'fragile' in _categories else 0):
                for c,v in iteritems(func.unspecified_categories):
                    setattr(func, c, v)
                    _categories.setdefault(c, v)
            default_updates = {}
            for c,v in iteritems(_categories):
                if c in func.unspecified_categories:
                    default_updates[c] = v
                setattr(func, c, v)
            if default_updates:
                for fcn in itervalues(func.__dict__):
                    if hasattr(fcn, '_categories'):
                        for c,v in iteritems(default_updates):
                            if c not in fcn._categories:
                                setattr(fcn, c, v)
        else:
            # This is a (currently unbound) method definition
            if len(_categories) > (1 if 'fragile' in _categories else 0):
                for c,v in iteritems(TestCase.unspecified_categories):
                    setattr(func, c, v)
            for c,v in iteritems(_categories):
                setattr(func, c, v)
            setattr(func, '_categories', _categories)
        return func
    return _id


#@nottest
def _run_import_baseline_test(self,
                              cwd=None,
                              module=None,
                              outfile=None,
                              baseline=None,
                              filter=None,
                              tolerance=None,
                              exact=False,
                              forceskip=False):
    if forceskip:
        if forceskip is True:
            self.skipTest("A forced test skip")
        else:
            self.skipTest(str(forceskip))
    if cwd is None:
        cwd = os.path.dirname(os.path.abspath(getfile(self.__class__)))
    oldpwd = os.getcwd()
    os.chdir(cwd)
    sys.path.insert(0, cwd)
    #
    try:
        import pyutilib.misc
        pyutilib.misc.setup_redirect(outfile)
        pyutilib.misc.import_file(module + ".py", clear_cache=True)
        pyutilib.misc.reset_redirect()
        #
        if baseline.endswith('.json'):
            self.assertMatchesJsonBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        elif baseline.endswith('.yml'):
            self.assertMatchesYamlBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        elif baseline.endswith('.xml'):
            self.assertMatchesXmlBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        else:
            self.assertFileEqualsBaseline(
                outfile, baseline, filter=filter, tolerance=tolerance)
    finally:
        os.chdir(oldpwd)
        sys.path.remove(cwd)


#@nottest
def _run_cmd_baseline_test(self,
                           cwd=None,
                           cmd=None,
                           outfile=None,
                           baseline=None,
                           filter=None,
                           cmdfile=None,
                           tolerance=None,
                           exact=False,
                           forceskip=False):
    if forceskip:
        if forceskip is True:
            self.skipTest("A forced test skip")
        else:
            self.skipTest(str(forceskip))
    if cwd is None:
        cwd = os.path.dirname(os.path.abspath(getfile(self.__class__)))
    oldpwd = os.getcwd()
    os.chdir(cwd)
    if type(cmd) in (list, tuple):
        useShell = False
    else:
        useShell = True
        cmd = cmd.strip()


    try:
        OUTPUT = open(outfile, "w")
        proc = subprocess.Popen(
            cmd, shell=useShell, stdout=OUTPUT, stderr=subprocess.STDOUT)
        proc.wait()
        OUTPUT.close()
        if cmdfile is not None:
            OUTPUT = open(cmdfile, 'w')
            OUTPUT.write("#!/bin/sh\n")
            OUTPUT.write("# Baseline test command\n")
            OUTPUT.write("#    cwd      %s\n" % cwd)
            OUTPUT.write("#    outfile  %s\n" % outfile)
            OUTPUT.write("#    baseline %s\n" % baseline)
            OUTPUT.write(cmd + '\n')
            OUTPUT.close()
            os.chmod(cmdfile, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
        if baseline.endswith('.json'):
            self.assertMatchesJsonBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        elif baseline.endswith('.yml'):
            self.assertMatchesYamlBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        elif baseline.endswith('.xml'):
            self.assertMatchesXmlBaseline(
                outfile, baseline, tolerance=tolerance, exact=exact)
        else:
            self.assertFileEqualsBaseline(
                outfile, baseline, filter=filter, tolerance=tolerance)
        if not cmdfile is None:
            os.remove(cmdfile)
    finally:
        OUTPUT.close()
        os.chdir(oldpwd)


#@nottest
def _run_fn_baseline_test(self,
                          fn=None,
                          name=None,
                          baseline=None,
                          filter=None,
                          tolerance=None):
    files = fn(name)
    self.assertFileEqualsBaseline(
        files[0], baseline, filter=filter, tolerance=tolerance)
    for file in files[1:]:
        os.remove(file)


#@nottest
def _run_fn_test(self, fn, name, suite):
    if suite is None:
        explanation = fn(self, name)
    else:
        explanation = fn(self, name, suite)
    if not explanation is None and explanation != "":
        self.fail(explanation)


class TestCase(unittest.TestCase):
    # Dictionary of options that may be used by function tests.
    _options = {}
    # The default test categories are 'smoke' and 'nightly' and 'expensive'
    smoke = 1
    nightly = 1
    expensive = 1
    fragile = 0
    _default_categories = True
    # If someone specifies a category, these are the default values of
    # the default categories if the category() decorator omits them.
    unspecified_categories = {
        'smoke':0, 'nightly':0, 'expensive':0, 'fragile':0 }

    def __init__(self, methodName='runTest'):
        unittest.TestCase.__init__(self, methodName)

    @staticmethod
    def parse_categories(category_string):
        return tuple(
            tuple(_category_to_tuple(_cat) for _cat in _set.split(','))
            for _set in category_string.split()
        )

    @staticmethod
    def categories_to_string(categories):
        return ' '.join(','.join("%s=%s" % y for y in x) for x in categories)


    def get_options(self, name, suite=None):
        return self._options[suite, name]

    @nottest
    def recordTestData(self, name, value):
        """A method for recording data associated with a test.  This method is only
           meaningful when running this TestCase with 'nose', using the TestData plugin.
        """
        tmp = getattr(self, 'testdata', None)
        if not tmp is None:
            tmp[name] = value

    def assertMatchesXmlBaseline(self,
                                 testfile,
                                 baseline,
                                 delete=True,
                                 tolerance=0.0,
                                 exact=False):
        try:
            import pyutilib.misc
            pyutilib.misc.compare_xml_files(
                baseline, testfile, tolerance=tolerance, exact=exact)
            if delete:
                os.remove(testfile)
        except Exception:
            err = sys.exc_info()[1]
            self.fail("XML testfile does not match the baseline:\n   testfile="
                      + testfile + "\n   baseline=" + baseline + "\n" + str(
                          err))

    def assertMatchesYamlBaseline(self,
                                  testfile,
                                  baseline,
                                  delete=True,
                                  tolerance=0.0,
                                  exact=False):
        try:
            import pyutilib.misc
            pyutilib.misc.compare_yaml_files(
                baseline, testfile, tolerance=tolerance, exact=exact)
            if delete:
                os.remove(testfile)
        except Exception:
            err = sys.exc_info()[1]
            self.fail("YAML testfile does not match the baseline:\n   testfile="
                      + testfile + "\n   baseline=" + baseline + "\n" + str(
                          err))

    def assertMatchesJsonBaseline(self,
                                  testfile,
                                  baseline,
                                  delete=True,
                                  tolerance=0.0,
                                  exact=False):
        try:
            import pyutilib.misc
            pyutilib.misc.compare_json_files(
                baseline, testfile, tolerance=tolerance, exact=exact)
            if delete:
                os.remove(testfile)
        except Exception:
            err = sys.exc_info()[1]
            self.fail("JSON testfile does not match the baseline:\n   testfile="
                      + testfile + "\n   baseline=" + baseline + "\n" + str(
                          err))

    def assertFileEqualsBaseline(self,
                                 testfile,
                                 baseline,
                                 filter=None,
                                 delete=True,
                                 tolerance=None):
        import pyutilib.misc
        [flag, lineno, diffs] = pyutilib.misc.compare_file(
            testfile, baseline, filter=filter, tolerance=tolerance)
        if not flag:
            if delete:
                os.remove(testfile)
        else:  #pragma:nocover
            self.fail("Unexpected output difference at line " + str(
                lineno) + ":\n   testfile=" + testfile + "\n   baseline=" +
                      baseline + "\nDiffs:\n" + diffs)
        return [flag, lineno]

    def assertFileEqualsLargeBaseline(self, testfile, baseline, delete=True):
        import pyutilib.misc
        flag = pyutilib.misc.compare_large_file(testfile, baseline)
        if not flag:
            if delete:
                os.remove(testfile)
        else:  #pragma:nocover
            self.fail("Unexpected output difference:\n   testfile=" + testfile +
                      "\n   baseline=" + baseline)
        return flag

    def assertFileEqualsBinaryFile(self, testfile, baseline, delete=True):
        theSame = filecmp.cmp(testfile, baseline)
        if theSame:
            if delete:
                os.remove(testfile)
        else:  #pragma:nocover
            self.fail("Unexpected output difference:\n   testfile=" + testfile +
                      "\n   baseline=" + baseline)
        return theSame

    @nottest
    def add_fn_test(cls, name=None, suite=None, fn=None, options=None):
        if fn is None:
            print("ERROR: must specify the 'fn' option to define the test")
            return
        if name is None:
            print("ERROR: must specify the 'name' option to define the test")
            return
        tmp = name.replace("/", "_")
        tmp = tmp.replace("\\", "_")
        tmp = tmp.replace(".", "_")
        func = lambda self, c1=fn, c2=name, c3=suite: _run_fn_test(self, c1, c2, c3)
        func.__name__ = "test_" + str(tmp)
        func.__doc__ = "function test: "+func.__name__+ \
                       " ("+str(cls.__module__)+'.'+str(cls.__name__)+")"
        setattr(cls, "test_" + tmp, func)
        cls._options[suite, name] = options

    add_fn_test = classmethod(add_fn_test)

    @nottest
    def add_baseline_test(cls,
                          name=None,
                          cmd=None,
                          fn=None,
                          baseline=None,
                          filter=None,
                          cwd=None,
                          cmdfile=None,
                          tolerance=None,
                          outfile=None,
                          exact=False,
                          forceskip=False):
        if cmd is None and fn is None:
            print(
                "ERROR: must specify either the 'cmd' or 'fn' option to define how the output file is generated")
            return
        if name is None:
            print("ERROR: must specify the test name")
            return
        if baseline is None:
            baseline = os.path.abspath(name + ".txt")
        tmp = name.replace("/", "_")
        tmp = tmp.replace("\\", "_")
        tmp = tmp.replace(".", "_")
        if outfile is None:
            (dirname, basename) = os.path.split(baseline)
            if '.' in basename:
                outfile = os.path.join(dirname,
                                       basename.rpartition('.')[0] + '.out')
            else:
                outfile = baseline + '.out'
        #
        # Create an explicit function so we can assign it a __name__ attribute.
        # This is needed by the 'nose' package
        #
        if fn is None:
            func = lambda self,c1=cwd,c2=cmd,c3=os.path.abspath(outfile),c4=baseline,c5=filter,c6=cmdfile,c7=tolerance,c8=exact,c9=forceskip: _run_cmd_baseline_test(self,cwd=c1,cmd=c2,outfile=c3,baseline=c4,filter=c5,cmdfile=c6,tolerance=c7,exact=c8,forceskip=c9)
        else:
            func = lambda self,c1=fn,c2=name,c3=baseline,c4=filter,c5=tolerance,c6=forceskip: _run_fn_baseline_test(self,fn=c1,name=c2,baseline=c3,filter=c4,tolerance=c5,forceskip=c6)
        func.__name__ = "test_" + tmp
        func.__doc__ = "baseline test: "+func.__name__+ \
                       " ("+str(cls.__module__)+'.'+str(cls.__name__)+")"
        if fn is None and not cmdfile is None:
            func.__doc__ += "  Command archived in " + os.path.abspath(cmdfile)
        setattr(cls, "test_" + tmp, func)

    add_baseline_test = classmethod(add_baseline_test)

    @nottest
    def add_import_test(cls,
                        module=None,
                        name=None,
                        cwd=None,
                        baseline=None,
                        filter=None,
                        tolerance=None,
                        outfile=None):
        if module is None and name is None:
            print("ERROR: must specify the module that is imported")
            return
        if module is None:
            module = name
        if name is None:
            print("ERROR: must specify test name")
            return
        if baseline is None:
            baseline = name + ".txt"
        tmp = name.replace("/", "_")
        tmp = tmp.replace("\\", "_")
        tmp = tmp.replace(".", "_")
        if outfile is None:
            (dirname, basename) = os.path.split(baseline)
            if '.' in basename:
                outfile = os.path.join(dirname,
                                       basename.rpartition('.')[0] + '.out')
            else:
                outfile = baseline + '.out'
        #
        # Create an explicit function so we can assign it a __name__ attribute.
        # This is needed by the 'nose' package
        #
        func = lambda self,c1=cwd,c2=module,c3=outfile,c4=baseline,c5=filter,c6=tolerance: _run_import_baseline_test(self,cwd=c1,module=c2,outfile=c3,baseline=c4,filter=c5,tolerance=c6)
        func.__name__ = "test_" + tmp
        func.__doc__ = "import test: " + func.__name__
        setattr(cls, "test_" + tmp, func)

    add_import_test = classmethod(add_import_test)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import re
from sys import exc_info, stdout
from textwrap import wrap
import logging
import pickle

import six
from six.moves import xrange

try:
    from yaml import dump
except ImportError:
    #dump = lambda x,**y: str(x)
    # YAML uses lowercase True/False
    def dump(x, **args):
        if type(x) is bool:
            return str(x).lower()
        return str(x)


try:
    import argparse
    argparse_is_available = True
except ImportError:
    argparse_is_available = False

__all__ = ('ConfigBlock', 'ConfigList', 'ConfigValue')

logger = logging.getLogger('pyutilib.misc')


def _munge_name(name, space_to_dash=True):
    if space_to_dash:
        name = re.sub(r'\s', '-', name)
    name = re.sub(r'_', '-', name)
    return re.sub(r'[^a-zA-Z0-9-_]', '_', name)


_leadingSpace = re.compile('^([ \n\t]*)')


def _strip_indentation(doc):
    if not doc:
        return doc
    lines = doc.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    if len(lines) == 1:
        return doc.lstrip()
    minIndent = min(len(_leadingSpace.match(l).group(0)) for l in lines[1:])
    if len(_leadingSpace.match(lines[0]).group(0)) <= minIndent:
        lines[0] = lines[0].strip()
    else:
        lines[0] = lines[0][minIndent:].rstrip()
    for i, l in enumerate(lines[1:]):
        lines[i + 1] = l[minIndent:].rstrip()
    return '\n'.join(lines)


class _UnpickleableDomain(object):
    def __init__(self, obj):
        self._type = type(obj).__name__
        self._name = obj.name(True)

    def __call__(self, arg):
        logging.error(
"""%s %s was pickled with an unpicklable domain.
    The domain was stripped and lost during the pickle process.  Setting
    new values on the restored object cannot be mapped into the correct
    domain.
""" % ( self._type, self._name))
        return arg

def _picklable(field,obj):
    try:
        pickle.dumps(field)
        return field
    except:
        # Contrary to the documentation, Python is not at all consistent
        # with the exception that is raised when pickling an object
        # fails:
        #
        #    Python 2.6 - 3.4:  pickle.PicklingError
        #    Python 3.5 - 4.6:  AttributeError
        #    Python 2.6 - 2.7 (cPickle):  TypeError
        #
        # What we are concerned about is masking things like recursion
        # errors.  Unfortunately, Python is not quite consistent there,
        # either: exceeding recursion depth raises a RuntimeError
        # through 3.4, then switches to a RecursionError (a derivative
        # of RuntimeError).
        if isinstance(exc_info()[0], RuntimeError):
            raise
        return _UnpickleableDomain(obj)


class ConfigBase(object):
    __slots__ = ('_parent', '_name', '_userSet', '_userAccessed', '_data',
                 '_default', '_domain', '_description', '_doc', '_visibility',
                 '_argparse')

    # This just needs to be any singleton-like object; we use it so that
    # we can tell if an argument is provided (and we can't use None as
    # None is a valid user-specified argument).  Making it a class helps
    # when Config objects are pickled.
    class NoArgument(object): pass

    def __init__(self,
                 default=None,
                 domain=None,
                 description=None,
                 doc=None,
                 visibility=0):
        self._parent = None
        self._name = None
        self._userSet = False
        self._userAccessed = False

        self._data = None
        self._default = default
        self._domain = domain
        self._description = _strip_indentation(description)
        self._doc = _strip_indentation(doc)
        self._visibility = visibility
        self._argparse = None

    def __getstate__(self):
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__slots__:
        #    state[i] = getattr(self,i)
        # return state
        #
        # Hoewever, in this case, the (nominal) parent class is
        # 'object', and object does not implement __getstate__.  Since
        # super() doesn't actually return a class, we are going to check
        # the *derived class*'s MRO and see if this is the second to
        # last class (the last is always 'object').  If it is, then we
        # can allocate the state dictionary.  If it is not, then we call
        # the super-class's __getstate__ (since that class is NOT
        # 'object').
        if self.__class__.__mro__[-2] is ConfigBase:
            state = {}
        else:
            state = super(ConfigBase, self).__getstate__()
        state.update((key, getattr(self, key)) for key in ConfigBase.__slots__)
        state['_domain'] = _picklable(state['_domain'], self)
        state['_parent'] = None
        return state

    def __setstate__(self, state):
        for key, val in six.iteritems(state):
            # Note: per the Python data model docs, we explicitly
            # set the attribute using object.__setattr__() instead
            # of setting self.__dict__[key] = val.
            object.__setattr__(self, key, val)

    def __call__(self, value=NoArgument):
        ans = self.__class__()
        ans._default     = self._default
        ans._domain      = self._domain
        ans._description = self._description
        ans._doc         = self._doc
        ans._visibility  = self._visibility
        if self.__class__ is ConfigBlock:
            ans._implicit_declaration = self._implicit_declaration
            ans._implicit_domain = self._implicit_domain
            for k in self._decl_order:
                if k in self._declared:
                    v = self._data[k]
                    ans._data[k] = _tmp = v()
                    ans._decl_order.append(k)
                    ans._declared.add(k)
                    _tmp._parent = ans
                    _tmp._name = v._name
        else:
            ans.reset()
        if value is not ConfigBase.NoArgument:
            ans.set_value(value)
        return ans

    def name(self, fully_qualified=False):
        # Special case for the top-level block
        if self._name is None:
            return ""
        elif fully_qualified and self._parent is not None:
            pName = self._parent.name(fully_qualified)
            # Special case for ConfigList indexing and the top-level entries
            if self._name.startswith('[') or not pName:
                return pName + self._name
            else:
                return pName + '.' + self._name
        else:
            return self._name

    def _cast(self, value):
        if value is None:
            return value
        if self._domain is not None:
            try:
                return self._domain(value)
            except:
                err = exc_info()[1]
                if hasattr(self._domain, '__name__'):
                    _dom = self._domain.__name__
                else:
                    _dom = type(self._domain)
                raise ValueError("invalid value for configuration '%s':\n"
                                 "\tFailed casting %s\n\tto %s\n\tError: %s" %
                                 (self.name(True), value, _dom, err))
        else:
            return value

    def reset(self):
        #
        # This is a dangerous construct, the failure in the first try block
        # can mask a real problem.
        #
        try:
            self.set_value(self._default)
        except:
            if hasattr(self._default, '__call__'):
                self.set_value(self._default())
            else:
                raise
        self._userAccessed = False
        self._userSet = False

    def declare_as_argument(self, *args, **kwds):
        """Map this Config item to an argparse argument.

Valid arguments include all valid arguments to argparse's
ArgumentParser.add_argument() with the exception of 'default'.  In addition,
you may provide a group keyword argument can be used to either pass in a
pre-defined option group or subparser, or else pass in the title of a
group, subparser, or (subparser, group)."""

        if 'default' in kwds:
            raise TypeError(
                "You cannot specify an argparse default value with "
                "ConfigBase.declare_as_argument().  The default value is "
                "supplied automatically from the Config definition.")

        if 'action' not in kwds and self._domain is bool:
            if not self._default:
                kwds['action'] = 'store_true'
            else:
                kwds['action'] = 'store_false'
                if not args:
                    args = ('--disable-' + _munge_name(self.name()),)
                if 'help' not in kwds:
                    kwds['help'] = "[DON'T] " + self._description
        if 'help' not in kwds:
            kwds['help'] = self._description
        if not args:
            args = ('--' + _munge_name(self.name()),)
        if self._argparse:
            self._argparse = self._argparse + ((args, kwds),)
        else:
            self._argparse = ((args, kwds),)
        return self

    def initialize_argparse(self, parser):

        def _get_subparser_or_group(_parser, name):
            # Note: strings also have a 'title()' method.  We are
            # looking for things that look like argparse
            # groups/subparsers, so just checking for the attribute
            # is insufficient: it needs to be a string attribute as
            # well
            if isinstance(name, argparse._ActionsContainer):
                #hasattr(_group, 'title') and \
                #    isinstance(_group.title, six.string_types):
                return 2, name

            if not isinstance(name, six.string_types):
                raise RuntimeError(
                    'Unknown datatype (%s) for argparse group on '
                    'configuration definition %s' %
                    (type(name).__name__, obj.name(True)))

            try:
                for _grp in _parser._subparsers._group_actions:
                    if name in _grp._name_parser_map:
                        return 1, _grp._name_parser_map[name]
            except AttributeError:
                pass

            for _grp in _parser._action_groups:
                if _grp.title == name:
                    return 0, _grp
            return 0, _parser.add_argument_group(title=name)

        def _process_argparse_def(_args, _kwds):
            _parser = parser
            # shallow copy the dict so we can remove the group flag and
            # add things like documentation, etc.
            _kwds = dict(_kwds)
            if 'group' in _kwds:
                _group = _kwds.pop('group')
                if isinstance(_group, tuple):
                    for _idx, _grp in enumerate(_group):
                        _issub, _parser = _get_subparser_or_group(_parser, _grp)
                        if not _issub and _idx < len(_group) - 1:
                            raise RuntimeError(
                                "Could not find argparse subparser '%s' for "
                                "Config item %s" % (_grp, obj.name(True)))
                else:
                    _issub, _parser = _get_subparser_or_group(_parser, _group)
            if 'dest' not in _kwds:
                _kwds['dest'] = 'CONFIGBLOCK.' + obj.name(True)
                if 'metavar' not in _kwds and \
                   _kwds.get('action','') not in ('store_true','store_false'):
                    if obj._domain is not None and \
                       obj._domain.__class__ is type:
                        _kwds['metavar'] = obj._domain.__name__.upper()
                    else:
                        _kwds['metavar'] = _munge_name(self.name().upper(),
                                                       False)
            _parser.add_argument(*_args, default=argparse.SUPPRESS, **_kwds)

        assert (argparse_is_available)
        for level, value, obj in self._data_collector(None, ""):
            if obj._argparse is None:
                continue
            for _args, _kwds in obj._argparse:
                _process_argparse_def(_args, _kwds)

    def import_argparse(self, parsed_args):
        for level, value, obj in self._data_collector(None, ""):
            if obj._argparse is None:
                continue
            for _args, _kwds in obj._argparse:
                if 'dest' in _kwds:
                    _dest = _kwds['dest']
                    if _dest in parsed_args:
                        obj.set_value(parsed_args.__dict__[_dest])
                else:
                    _dest = 'CONFIGBLOCK.' + obj.name(True)
                    if _dest in parsed_args:
                        obj.set_value(parsed_args.__dict__[_dest])
                        del parsed_args.__dict__[_dest]
        return parsed_args

    def display(self, content_filter=None, indent_spacing=2, ostream=None):
        if content_filter not in ConfigBlock.content_filters:
            raise ValueError("unknown content filter '%s'; valid values are %s"
                             % (content_filter, ConfigBlock.content_filters))

        _blocks = []
        if ostream is None:
            ostream=stdout

        for level, value, obj in self._data_collector(0, ""):
            if content_filter == 'userdata' and not obj._userSet:
                continue

            _blocks[level:] = [' ' * indent_spacing * level + value + "\n",]

            for i, v in enumerate(_blocks):
                if v is not None:
                    ostream.write(v)
                    _blocks[i] = None

    def generate_yaml_template(self, indent_spacing=2, width=78, visibility=0):
        minDocWidth = 20
        comment = "  # "
        data = list(self._data_collector(0, "", visibility))
        level_info = {}
        for lvl, val, obj in data:
            if lvl not in level_info:
                level_info[lvl] = {'data': [], 'off': 0, 'line': 0, 'over': 0}
            level_info[lvl]['data'].append(
                (val.find(':') + 2, len(val), len(obj._description or "")))
        for lvl in sorted(level_info):
            indent = lvl * indent_spacing
            _ok = width - indent - len(comment) - minDocWidth
            offset = \
                max( val if val < _ok else key
                     for key,val,doc in level_info[lvl]['data'] )
            offset += indent + len(comment)
            over = sum(1 for key, val, doc in level_info[lvl]['data']
                       if doc + offset > width)
            if len(level_info[lvl]['data']) - over > 0:
                line = max(offset + doc
                           for key, val, doc in level_info[lvl]['data']
                           if offset + doc <= width)
            else:
                line = width
            level_info[lvl]['off'] = offset
            level_info[lvl]['line'] = line
            level_info[lvl]['over'] = over
        maxLvl = 0
        maxDoc = 0
        pad = 0
        for lvl in sorted(level_info):
            _pad = level_info[lvl]['off']
            _doc = level_info[lvl]['line'] - _pad
            if _pad > pad:
                if maxDoc + _pad <= width:
                    pad = _pad
                else:
                    break
            if _doc + pad > width:
                break
            if _doc > maxDoc:
                maxDoc = _doc
            maxLvl = lvl
        os = six.StringIO()
        if self._description:
            os.write(comment.lstrip() + self._description + "\n")
        for lvl, val, obj in data:
            if not obj._description:
                os.write(' ' * indent_spacing * lvl + val + "\n")
                continue
            if lvl <= maxLvl:
                field = pad - len(comment)
            else:
                field = level_info[lvl]['off'] - len(comment)
            os.write(' ' * indent_spacing * lvl)
            if width - len(val) - minDocWidth >= 0:
                os.write('%%-%ds' % (field - indent_spacing * lvl) % val)
            else:
                os.write(val + '\n' + ' ' * field)
            os.write(comment)
            txtArea = max(width - field - len(comment), minDocWidth)
            os.write(("\n" + ' ' * field + comment).join(
                wrap(
                    obj._description, txtArea, subsequent_indent='  ')))
            os.write('\n')
        return os.getvalue()

    def generate_documentation\
            ( self,
              block_start= "\\begin{description}[topsep=0pt,parsep=0.5em,itemsep=-0.4em]\n",
              block_end=   "\\end{description}\n",
              item_start=  "\\item[{%s}]\\hfill\n",
              item_body=   "\\\\%s",
              item_end=    "",
              indent_spacing=2,
              width=78,
              visibility=0
              ):
        os = six.StringIO()
        level = []
        lastObj = self
        indent = ''
        for lvl, val, obj in self._data_collector(1, '', visibility, True):
            #print len(level), lvl, val, obj
            if len(level) < lvl:
                while len(level) < lvl - 1:
                    level.append(None)
                level.append(lastObj)
                if '%s' in block_start:
                    os.write(indent + block_start % lastObj.name())
                elif block_start:
                    os.write(indent + block_start)
                indent += ' ' * indent_spacing
            while len(level) > lvl:
                _last = level.pop()
                if _last is not None:
                    indent = indent[:-1 * indent_spacing]
                    if '%s' in block_end:
                        os.write(indent + block_end % _last.name())
                    elif block_end:
                        os.write(indent + block_end)

            lastObj = obj
            if '%s' in item_start:
                os.write(indent + item_start % obj.name())
            elif item_start:
                os.write(indent + item_start)
            _doc = obj._doc or obj._description or ""
            if _doc:
                _wrapLines = '\n ' not in _doc
                if '%s' in item_body:
                    _doc = item_body % (_doc,)
                elif _doc:
                    _doc = item_body
                if _wrapLines:
                    doc_lines = wrap(
                        _doc,
                        width,
                        initial_indent=indent + ' ' * indent_spacing,
                        subsequent_indent=indent + ' ' * indent_spacing)
                else:
                    doc_lines = (_doc,)
                # Write things out
                os.writelines('\n'.join(doc_lines))
                if not doc_lines[-1].endswith("\n"):
                    os.write('\n')
            if '%s' in item_end:
                os.write(indent + item_end % obj.name())
            elif item_end:
                os.write(indent + item_end)
        while level:
            indent = indent[:-1 * indent_spacing]
            _last = level.pop()
            if '%s' in block_end:
                os.write(indent + block_end % _last.name())
            else:
                os.write(indent + block_end)
        return os.getvalue()

    def user_values(self):
        if self._userSet:
            yield self
        for level, value, obj in self._data_collector(0, ""):
            if obj._userSet:
                yield obj

    def unused_user_values(self):
        if self._userSet and not self._userAccessed:
            yield self
        for level, value, obj in self._data_collector(0, ""):
            if obj._userSet and not obj._userAccessed:
                yield obj


class ConfigValue(ConfigBase):

    def __init__(self, *args, **kwds):
        ConfigBase.__init__(self, *args, **kwds)
        self.reset()

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return self._data

    def set_value(self, value):
        self._data = self._cast(value)
        self._userSet = True

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        _str = dump(self._data, default_flow_style=True).rstrip()
        if _str.endswith("..."):
            _str = _str[:-3].rstrip()
        yield (level, prefix + _str, self)


class ConfigList(ConfigBase):

    def __init__(self, *args, **kwds):
        ConfigBase.__init__(self, *args, **kwds)
        if self._domain is None:
            self._domain = ConfigValue()
        elif isinstance(self._domain, ConfigBase):
            pass
        else:
            self._domain = ConfigValue(None, domain=self._domain)
        self.reset()


    def __setstate__(self, state):
        state = super(ConfigList, self).__setstate__(state)
        for x in self._data:
            x._parent = self

    def __getitem__(self, key):
        val = self._data[key]
        self._userAccessed = True
        if isinstance(val, ConfigValue):
            return val.value()
        else:
            return val

    def get(self, key, default=ConfigBase.NoArgument):
        # Note: get() is borrowed from ConfigBlock for cases where we
        # want the raw stored object (and to aviod the implicit
        # conversion of ConfigValue members to their stored data).
        try:
            val = self._data[key]
            self._userAccessed = True
            return val
        except:
            pass
        if default is ConfigBase.NoArgument:
            return None
        if self._domain is not None:
            return self._domain(default)
        else:
            return ConfigValue(default)

    def __setitem__(self, key, val):
        # Note: this will fail if the element doesn't exist in _data.
        # As a result, *this* list doesn't change when someone tries to
        # change an element; instead, the *element* gets its _userSet
        # flag set.
        #self._userSet = True
        self._data[key].set_value(val)

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        self._userAccessed = True
        return iter(self[i] for i in xrange(len(self._data)))

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return [config.value(accessValue) for config in self._data]

    def set_value(self, value):
        # If the set_value fails part-way through the list values, we
        # want to restore a deterministic state.  That is, either
        # set_value succeeds completely, or else nothing happens.
        _old = self._data
        self._data = []
        try:
            if (type(value) is list) or \
               isinstance(value, ConfigList):
                for val in value:
                    self.append(val)
            else:
                self.append(value)
        except:
            self._data = _old
            raise
        self._userSet = True

    def reset(self):
        ConfigBase.reset(self)
        # Because the base reset() calls set_value, any deefault list
        # entries will get their userSet flag set.  This is wrong, as
        # reset() should conceptually reset teh object to it's default
        # state (e.g., before the user ever had a chance to mess with
        # things).  As the list could contain a ConfigBlock, this is a
        # recursive operation to put the userSet values back.
        for val in self.user_values():
            val._userSet = False

    def append(self, value=ConfigBase.NoArgument):
        val = self._cast(value)
        if val is None:
            return
        self._data.append(val)
        #print self._data[-1], type(self._data[-1])
        self._data[-1]._parent = self
        self._data[-1]._name = '[%s]' % (len(self._data) - 1,)
        self._data[-1]._userSet = True
        self._userSet = True

    def add(self, value=ConfigBase.NoArgument):
        logger.warning(
            "DEPRECATED: ConfigList.add() has been deprecated.  Use append()")
        return self.append(value)

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        if docMode:
            # In documentation mode, we do NOT list the documentation
            # for any sub-data, and instead document the *domain*
            # information (as all the entries should share the same
            # domain, potentially duplicating that documentation is
            # somewhat redundant, and worse, if the list is empty, then
            # no documentation is generated at all!)
            yield (level, prefix.rstrip(), self)
            subDomain = self._domain._data_collector(level + 1, '- ',
                                                     visibility, docMode)
            # Pop off the (empty) block entry
            six.next(subDomain)
            for v in subDomain:
                yield v
            return
        if prefix:
            if not self._data:
                yield (level, prefix.rstrip() + ' []', self)
            else:
                yield (level, prefix.rstrip(), self)
                if level is not None:
                    level += 1
        for value in self._data:
            for v in value._data_collector(level, '- ', visibility, docMode):
                yield v


class ConfigBlock(ConfigBase):
    content_filters = (None, 'all', 'userdata')

    __slots__ = ('_decl_order', '_declared', '_implicit_declaration',
                 '_implicit_domain')
    _all_slots = __slots__ + ConfigBase.__slots__

    def __init__(self,
                 description=None,
                 doc=None,
                 implicit=False,
                 implicit_domain=None,
                 visibility=0):
        self._decl_order = []
        self._declared = set()
        self._implicit_declaration = implicit
        if implicit_domain is None or isinstance(implicit_domain, ConfigBase):
            self._implicit_domain = implicit_domain
        else:
            self._implicit_domain = ConfigValue(None, domain=implicit_domain)
        ConfigBase.__init__(self, None, {}, description, doc, visibility)
        self._data = {}

    def __getstate__(self):
        state = super(ConfigBlock, self).__getstate__()
        state.update((key, getattr(self, key)) for key in ConfigBlock.__slots__)
        state['_implicit_domain'] = _picklable(state['_implicit_domain'], self)
        return state

    def __setstate__(self, state):
        state = super(ConfigBlock, self).__setstate__(state)
        for x in six.itervalues(self._data):
            x._parent = self

    def __getitem__(self, key):
        self._userAccessed = True
        key = str(key)
        if isinstance(self._data[key], ConfigValue):
            return self._data[key].value()
        else:
            return self._data[key]

    def get(self, key, default=ConfigBase.NoArgument):
        self._userAccessed = True
        key = str(key)
        if key in self._data:
            return self._data[key]
        if default is ConfigBase.NoArgument:
            return None
        if self._implicit_domain is not None:
            return self._implicit_domain(default)
        else:
            return ConfigValue(default)

    def setdefault(self, key, default=ConfigBase.NoArgument):
        self._userAccessed = True
        key = str(key)
        if key in self._data:
            return self._data[key]
        if default is ConfigBase.NoArgument:
            return self.add(key, None)
        else:
            return self.add(key, default)

    def __setitem__(self, key, val):
        key = str(key)
        if key not in self._data:
            self.add(key, val)
        else:
            self._data[key].set_value(val)
        #self._userAccessed = True

    def __contains__(self, key):
        key = str(key)
        return key in self._data

    def __len__(self):
        return self._decl_order.__len__()

    def __iter__(self):
        return self._decl_order.__iter__()

    def __getattr__(self, name):
        # Note: __getattr__ is only called after all "usual" attribute
        # lookup methods have failed.  So, if we get here, we already
        # know that key is not a __slot__ or a method, etc...
        #if name in ConfigBlock._all_slots:
        #    return super(ConfigBlock,self).__getattribute__(name)
        if name not in self._data:
            _name = name.replace('_', ' ')
            if _name not in self._data:
                raise AttributeError("Unknown attribute '%s'" % name)
            name = _name
        return ConfigBlock.__getitem__(self, name)

    def __setattr__(self, name, value):
        if name in ConfigBlock._all_slots:
            super(ConfigBlock, self).__setattr__(name, value)
        else:
            if name not in self._data:
                name = name.replace('_', ' ')
            ConfigBlock.__setitem__(self, name, value)

    def iterkeys(self):
        return self._decl_order.__iter__()

    def itervalues(self):
        self._userAccessed = True
        for key in self._decl_order:
            yield self[key]

    def iteritems(self):
        self._userAccessed = True
        for key in self._decl_order:
            yield (key, self[key])

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def _add(self, name, config):
        name = str(name)
        if config._parent is not None:
            raise ValueError(
                "config '%s' is already assigned to Config Block '%s'; "
                "cannot reassign to '%s'" %
                (name, config._parent.name(True), self.name(True)))
        if name in self._data:
            raise ValueError(
                "duplicate config '%s' defined for Config Block '%s'" %
                (name, self.name(True)))
        if '.' in name or '[' in name or ']' in name:
            raise ValueError(
                "Illegal character in config '%s' for config Block '%s': "
                "'.[]' are not allowed." % (name, self.name(True)))
        self._data[name] = config
        self._decl_order.append(name)
        config._parent = self
        config._name = name
        return config

    def declare(self, name, config):
        ans = self._add(name, config)
        self._declared.add(name)
        return ans

    def add(self, name, config):
        if not self._implicit_declaration:
            raise ValueError("Key '%s' not defined in Config Block '%s'"
                             " and Block disallows implicit entries" %
                             (name, self.name(True)))

        if self._implicit_domain is None:
            if isinstance(config, ConfigBase):
                ans = self._add(name, config)
            else:
                ans = self._add(name, ConfigValue(config))
        else:
            ans = self._add(name, self._implicit_domain(config))
        self._userSet = True
        return ans

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return dict((name, config.value(accessValue))
                    for name, config in six.iteritems(self._data))

    def set_value(self, value):
        if value is None:
            return self
        if (type(value) is not dict) and \
           (not isinstance(value, ConfigBlock)):
            raise ValueError("Expected dict value for %s.set_value, found %s" %
                             (self.name(True), type(value).__name__))
        if not value:
            return self
        _implicit = []
        _decl_map = {}
        for key in value:
            if str(key) in self._data:
                # str(key) may not be key... store the mapping so that
                # when we later iterate over the _decl_order, we can map
                # the local keys back to the incoming value keys.
                _decl_map[str(key)] = key
            else:
                if self._implicit_declaration:
                    _implicit.append(key)
                else:
                    raise ValueError(
                        "key '%s' not defined for Config Block '%s' and "
                        "implicit (undefined) keys are not allowed" %
                        (key, self.name(True)))

        # If the set_value fails part-way through the new values, we
        # want to restore a deterministic state.  That is, either
        # set_value succeeds completely, or else nothing happens.
        _old_data = self.value(False)
        try:
            # We want to set the values in declaration order (so that
            # things are deterministic and in case a validation depends
            # on the order)
            for key in self._decl_order:
                if key in _decl_map:
                    #print "Setting", key, " = ", value
                    self._data[key].set_value(value[_decl_map[key]])
            # implicit data is declared at the end (in sorted order)
            for key in sorted(_implicit):
                self.add(key, value[key])
        except:
            self.reset()
            self.set_value(_old_data)
            raise
        self._userSet = True
        return self

    def reset(self):
        # Reset the values in the order they were declared.  This
        # allows reset functions to have a deterministic ordering.
        def _keep(self, key):
            keep = key in self._declared
            if keep:
                self._data[key].reset()
            else:
                del self._data[key]
            return keep
        # this is an in-place slice of a list...
        self._decl_order[:] = [x for x in self._decl_order if _keep(self, x)]
        self._userAccessed = False
        self._userSet = False

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        if prefix:
            yield (level, prefix.rstrip(), self)
            if level is not None:
                level += 1
        for key in self._decl_order:
            for v in self._data[key]._data_collector(level, key + ': ',
                                                     visibility, docMode):
                yield v

# In Python3, the items(), etc methods of dict-like things return
# generator-like objects.
if six.PY3:
    ConfigBlock.keys = ConfigBlock.iterkeys
    ConfigBlock.values = ConfigBlock.itervalues
    ConfigBlock.items = ConfigBlock.iteritems



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import fnmatch
import errno
import linecache
import os
import re
import shutil
import stat
import sys
import warnings

if (sys.platform[0:3] == "win"):  #pragma:nocover
    executable_extension = ".exe"
else:  #pragma:nocover
    executable_extension = ""

import six


def deprecated(deprecated_function):
    """ Code slightly adapted from the Python Decorator Library

    This is a decorator which can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used.

    Example use:

    @deprecated
    def some_func ( ):
        ...
    """

    def wrapper_function(*args, **kwargs):
        warnings.warn_explicit(
            "Use of deprecated function '%s'." % deprecated_function.__name__,
            category=DeprecationWarning,
            filename=deprecated_function.__code__.co_filename,
            lineno=deprecated_function.__code__.co_firstlineno + 1)
        return deprecated_function(*args, **kwargs)

    return wrapper_function


def tostr(array):
    """ Create a string from an array of numbers """
    tmpstr = ""
    for val in array:
        tmpstr = tmpstr + " " + repr(val)
    return tmpstr.strip()


def flatten(x):
    """Flatten nested iterables"""

    def _flatten(x, ans_):
        for el in x:
            if not type(el) is str and hasattr(el, "__iter__"):
                # NB: isinstance can be SLOW if it is going to return false,
                # so we will do one extra hasattr() check that will pretty
                # much assure that it will be True
                # NB: we will flatten anything that looks iterable, except strings
                try:
                    el_it = iter(el)
                    _flatten(el_it, ans_)
                except:
                    ans_.append(el)
            else:
                ans_.append(el)

    # flatten() is really just a recursive routine; however, if we do a
    # naive recursive call, we end up creating small temporary lists and
    # throwing them away.  We could add an optional second argument
    # (like _flatten), but then we need to do an "if ans is None: ans =
    # []" check.  This dual-function approach appears to be the most
    # efficient.
    if type(x) is str or not hasattr(x, "__iter__"):
        return x
    ans = []
    _flatten(x, ans)
    return ans


def flatten_list(x):
    """Flatten nested lists"""
    if type(x) is not list:
        return x
    x_len = len(x)
    i = 0
    while i < x_len:
        if type(x[i]) is list:
            x_len += len(x[i]) - 1
            x[i:i + 1] = x[i]
        else:
            i += 1
    return x


def recursive_flatten_tuple(val):
    """ Flatten nested tuples """
    if type(val) is not tuple:
        return val
    rv = ()
    for i in val:
        if type(i) is tuple:
            rv = rv + flatten_tuple(i)
        else:
            rv = rv + (i,)
    return rv


def flatten_tuple(x):
    """ Flatten nested tuples """
    if type(x) is not tuple:
        return x
    x_len = len(x)
    i = 0
    while i < x_len:
        if type(x[i]) is tuple:
            x_len += len(x[i]) - 1
            x = x[:i] + x[i] + x[i + 1:]
        else:
            i += 1
    return x


#
# A better method for removing directory trees, which handles MSWindows errors
# that are associated with read-only files.
#
def handleRemoveReadonly(func, path, exc):
    excvalue = exc[1]
    if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


def rmtree(dir):
    if not os.path.exists(dir):
        return
    shutil.rmtree(dir, ignore_errors=False, onerror=handleRemoveReadonly)


whitespace_re = re.compile('\s+')


def quote_split(regex_str, src=None):
    """
    Split a string, but do not split the string between quotes.  If only
    one argument is provided (the string to be split), regex_str
    defaults to '\s+'.  In addition, regex_str can be either a regular
    expression string, or a compiled expression.
    """

    if src is None:
        src = regex_str
        regex = whitespace_re
    elif 'match' in dir(regex_str):
        regex = regex_str
    else:
        regex = re.compile(regex_str)

    # We need to figure out where the quoted strings are.  Given that
    # lots of things may be escaped (e.g., '\\\"'), we can only find the
    # "real" quotes by walking the entire string. <sigh>.
    tokens = []
    start = 0
    inQuote = ''
    escaping = False
    for idx, char in enumerate(src):
        if escaping:
            escaping = False
        elif char == inQuote:
            inQuote = ''
        else:
            if not inQuote and start <= idx:
                g = regex.match(src[idx:])
                if g:
                    tokens.append(src[start:idx])
                    start = idx + len(g.group())
                    # NB: we still want to parse the remainder of the patern
                    # for things like escape characters so that we correctly
                    # parse the entire source string; hence, no 'elif'
            if char == '\\':
                escaping = True
            elif not inQuote and (char == '"' or char == "'"):
                inQuote = char

    if inQuote:
        raise ValueError("ERROR: unterminated quotation found in quote_split()")

    tokens.append(src[start:])
    return tokens


def traceit(frame, event, arg):  #pragma:nocover
    """
    A utility for tracing Python executions.  Use this function by
    executing:

    sys.settrace(traceit)
    """
    if event == "line":
        lineno = frame.f_lineno
        try:
            filename = frame.f_globals["__file__"]
        except:
            return traceit
        if (filename.endswith(".pyc") or filename.endswith(".pyo")):
            filename = filename[:-1]
        name = frame.f_globals["__name__"]
        line = linecache.getline(filename, lineno)
        print("%s:%s: %s" % (name, lineno, line.rstrip()))
    return traceit


def tuplize(dlist, d, name):
    """
    Convert a list into a list of tuples.
    """
    if len(dlist) % d != 0:
        raise ValueError("Cannot tuplize data for set " + str(
            name) + " because its length " + str(len(dlist)) +
                         " is not a multiple of dimen " + str(d))
    j = 0
    t = []
    rv = []
    for i in dlist:
        t.append(i)
        j += 1
        if j == d:
            rv.append(tuple(t))
            t = []
            j = 0
    return rv


def find_files(directory, *args):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            for pattern in args:
                if fnmatch.fnmatch(basename, pattern):
                    filename = os.path.join(root, basename)
                    yield filename


def search_file(filename,
                search_path=None,
                implicitExt=executable_extension,
                executable=False,
                isfile=True,
                validate=None):
    """
    Given a search path, find a file.

    Can specify the following options:
       search_path - A list of directories that are searched, or the
           name of a single directory to search. If unspecified,
           then all directories in the PATH defined for the current
           environment will be searched.
       executable_extension - This string is used to see if there is an
           implicit extension in the filename
       executable - Test if the file is an executable (default=False)
       isfile - Test if the file is a file (default=True)
    """
    if search_path is None:
        #
        # Use the PATH environment if it is defined and not empty
        #
        if "PATH" in os.environ and os.environ["PATH"] != "":
            search_path = os.environ['PATH'].split(os.pathsep)
        else:
            search_path = os.defpath.split(os.pathsep)
    else:
        if isinstance(search_path, six.string_types):
            search_path = (search_path,)
    for path in search_path:
        for ext in ('', implicitExt):
            test_fname = os.path.join(path, filename + ext)
            if os.path.exists(test_fname) \
                   and (not isfile or os.path.isfile(test_fname)) \
                   and (not executable or os.access(test_fname, os.X_OK)):
                file = os.path.abspath(test_fname)
                if validate is None or validate(file):
                    return file
    return None


def sort_index(l):
    """Returns a list, where the i-th value is the index of the i-th smallest
    value in the data 'l'"""
    return list(index
                for index, item in sorted(
                    enumerate(l), key=lambda item: item[1]))


def count_lines(file):
    """Returns the number of lines in a file."""
    count = 0
    for line in open(file, "r"):
        count = count + 1
    return count


class Bunch(dict):
    """
    A class that can be used to store a bunch of data dynamically

    foo = Bunch(data=y, sq=y*y, val=2)
    print foo.data
    print foo.sq
    print foo.val

    Adapted from code developed by Alex Martelli and submitted to
    the ActiveState Programmer Network http://aspn.activestate.com
    """

    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__.update(kw)


class Container(dict):
    """
    A generalization of Bunch.  This class allows all other attributes to have a
    default value of None.  This borrows the output formatting ideas from the
    ActiveState Code Container (recipe 496697).
    """

    def __init__(self, *args, **kw):
        for arg in args:
            for item in quote_split('[ \t]+', arg):
                r = item.find('=')
                if r != -1:
                    try:
                        val = eval(item[r + 1:])
                    except:
                        val = item[r + 1:]
                    kw[item[:r]] = val
        dict.__init__(self, kw)
        self.__dict__.update(kw)
        if not '_name_' in kw:
            self._name_ = self.__class__.__name__

    def update(self, d):
        """
        The update is specialized for JSON-like data.  This
        recursively replaces dictionaries with Container objects.
        """
        for k in d:
            if type(d[k]) is dict:
                tmp = Container()
                tmp.update(d[k])
                self.__setattr__(k, tmp)
            elif type(d[k]) is list:
                val = []
                for i in d[k]:
                    if type(i) is dict:
                        tmp = Container()
                        tmp.update(i)
                        val.append(tmp)
                    else:
                        val.append(i)
                self.__setattr__(k, val)
            else:
                self.__setattr__(k, d[k])

    def set_name(self, name):
        self._name_ = name

    def __setitem__(self, name, val):
        self.__setattr__(name, val)

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __setattr__(self, name, val):
        if name[0] != '_':
            dict.__setitem__(self, name, val)
        self.__dict__[name] = val

    def __getattr__(self, name):
        try:
            return dict.__getitem__(self, name)
        except:
            if name[0] == '_':
                raise AttributeError("Unknown attribute %s" % name)
        return None

    def __repr__(self):
        attrs = sorted("%s = %r" % (k, v) for k, v in self.__dict__.items()
                       if not k.startswith("_"))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(attrs))

    def __str__(self):
        return self.as_string()

    def __str__(self, nesting=0, indent=''):
        attrs = []
        indentation = indent + "    " * nesting
        for k, v in self.__dict__.items():
            if not k.startswith("_"):
                text = [indentation, k, ":"]
                if isinstance(v, Container):
                    if len(v) > 0:
                        text.append('\n')
                    text.append(v.__str__(nesting + 1))
                elif isinstance(v, list):
                    if len(v) == 0:
                        text.append(' []')
                    else:
                        for v_ in v:
                            text.append('\n' + indentation + "-")
                            if isinstance(v_, Container):
                                text.append('\n' + v_.__str__(nesting + 1))
                            else:
                                text.append(" " + repr(v_))
                else:
                    text.append(' ' + repr(v))
                attrs.append("".join(text))
        attrs.sort()
        return "\n".join(attrs)


class Options(Container):
    """
    This is a convenience class.  A common use of the Container class is to
    manage options that are passed into a class/solver/framework.  Thus,
    it's convenient to call this an Options object.
    """

    def __init__(self, *args, **kw):
        Container.__init__(self, *args, **kw)
        self.set_name('Options')


def create_hardlink(src, dst):
    """
    Create a hard link where dst points to src.
    """
    if os.name == 'nt' and sys.version_info[:2] < (3,2):
        # Windows; Python added native support for hard links in 3.2
        import ctypes
        if not ctypes.windll.kernel32.CreateHardLinkA(dst, src, 0):
            raise OSError("Failed to create hard link for file %s" % (src))
    else:
        # Unix; Windows >= Python3.2
        os.link(src, dst)



"""This module defines a nose plugin to terminate a test after a
specified number of seconds.

Use the following command-line option with nosetests ::

    nosetests --test-timeout=###

"""

__all__ = ['Timeout', 'TestTimeout']

import os
import signal
from nose.plugins.base import Plugin
try:
    from psutil import Process
    _psutil_avail = True
except ImportError:
    _psutil_avail = False
except NotImplementedError:
    _psutil_avail = False


class Timeout(Exception):
    pass


class TestTimeout(Plugin):
    """Kill tests if they exceed the specified timeout."""
    name = 'timeout'
    score = 5000  # Run early

    def options(self, parser, env):
        """Register command-line options."""
        parser.add_option(
            "--test-timeout",
            action="store",
            default=env.get('NOSE_TEST_TIMEOUT', 0),
            dest="test_timeout",
            metavar="SECONDS",
            help="A per-test timeout (in seconds). "
            "[NOSE_TEST_TIMEOUT]")

    def configure(self, options, config):
        self.timeout = int(options.test_timeout)
        self.enabled = self.timeout > 0
        if self.enabled and not _psutil_avail:
            self.enabled = False
            raise ImportError(
                "The nose Timeout plugin requires the psutil package.")

    def startTest(self, test):
        signal.signal(signal.SIGALRM, self._killTest)
        signal.alarm(self.timeout)

    def stopTest(self, test):
        signal.alarm(0)

    def _all_children(self, p):
        ans = p.get_children()
        i = 0
        while i < len(ans):
            ans.extend(self._all_children(ans[i]))
            i += 1
        return ans

    def _killTest(self, signum, frame):
        for p in self._all_children(Process(os.getpid())):
            try:
                p.kill()
            except:
                pass
        hour = int(self.timeout / 3600)
        min = int(self.timeout / 60) - hour * 60
        sec = self.timeout % 60
        txt = ""
        if hour:
            txt = "%d hour%s" % (hour, hour > 1 and "s" or "")
        if min:
            if txt:
                txt += ", "
            txt += "%d minute%s" % (min, min > 1 and "s" or "")
        if sec:
            if txt:
                txt += ", "
            txt += "%d second%s" % (sec, sec > 1 and "s" or "")
        raise Timeout("Test exceeded timeout (%s)" % txt)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________


class Factory(object):
    """
    Class that provides a generic constructor mechanism.

    Adapted from code developed by Andres Tuells and submitted to
    the ActiveState Programmer Network http://aspn.activestate.com
    """

    def __init__(self):
        """ Constructor """
        self.Keys = []
        self.constructors = {}

    def register(self, methodName, constructor, *args, **kargs):
        """ Register a constructor """
        self.Keys = self.Keys + [methodName]
        _args = [constructor]
        _args.extend(args)
        mname = "construct_" + methodName
        self.constructors[methodName] = _Functor(*_args, **kargs)
        #self.constructors[methodName] = apply(_Functor,_args, kargs)
        setattr(self, methodName, _Functor(*_args, **kargs))
        #setattr(self, methodName,apply(_Functor,_args, kargs))

    def construct(self, name):
        """
        Use the string name to create an object.
        """
        try:
            method = getattr(self, str(name))
        except AttributeError:
            return None
        return method()

    def __call__(self, name):
        """
        Use the string name to create an object.
        """
        return self.construct(name)

    def unregister(self, methodName):
        """ Unregister a constructor """
        delattr(self, methodName)
        self.Keys.remove(methodName)

    def keys(self):
        """
        Returns the list of names that can are registered in this factory
        """
        return self.Keys

    def __len__(self):
        """
        Returns the number of names in the factory
        """
        return len(self.Keys)

    def __iter__(self):
        """ Returns an iterator for the factory keys """
        return self.Keys.__iter__()

    def __getitem__(self, index):
        """ Allows an index into the list of factory keys """
        return self.Keys[index]


class _Functor:
    """
    A class that wraps a function to act like the function.
    """

    def __init__(self, function, *args, **kargs):
        """ Constructor """
        assert hasattr(function,
                       '__call__'), "function should be a callable obj"
        self._function = function
        self._args = args
        self._kargs = kargs

    def __call__(self, *args, **kargs):
        """ Call function """
        _args = list(self._args)
        _args.extend(args)
        _kargs = self._kargs.copy()
        _kargs.update(kargs)
        return self._function(*_args, **_kargs)



from collections import deque


class SimpleVisitor(object):

    def visit(self, node):  #pragma: no cover
        """
	    Visit a node in a tree and perform some operation on
	    it.

        This method should be over-written by a user
        that is creating a sub-class.

        Args:
            node: a node in a tree

        Returns:
            nothing
        """
        pass

    def children(self, node):
        """
        Return the children for a node in a tree.
        
        This method has a default implementation, but
        this implementation will likely be redefined
        to reflect the tree data structures used for 
        a specific application.

        Args:
            node: a node in a tree

        Returns:
            A list of children of the specified node.
        """ 
        return node.children
    
    def is_leaf(self, node):
        """
        Return :const:`True` if the node has no children.

        Args:
            node: a node in a tree

        Returns:
            :const:`True` if this node has no children, and :const:`False`
            otherwise
        """
        return len(node.children) == 0

    def finalize(self):     #pragma: no cover
        """
        Return the "final value" of the search.

        The default implementation returns :const:`None`, because
        the traditional visitor pattern does not return a value.

        Returns:
            The final value after the search.  Default is :const:`None`.
        """
        pass

    def bfs(self, node):
        """
        Perform breadth-first search starting at a node.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        dq = deque([node])
        while dq:
            current = dq.popleft()
            self.visit(current)
            if not self.is_leaf(current):
                dq.extend(self.children(current))
        return self.finalize()

    def xbfs(self, node):
        """
        Perform breadth-first search starting at a node,
        except that leaf nodes are immediately visited.

        Immediately visiting leaf nodes eliminates 
        storing and retrieving leaf nodes, which is an expense
        that may not be necessary if the search order can be
        changed from the standard BFS.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        dq = deque([node])
        while dq:
            current = dq.popleft()
            self.visit(current)
            for c in self.children(current):
                if self.is_leaf(c):
                    self.visit(c)
                else:
                    dq.append(c)
        return self.finalize()

    def dfs_preorder(self, node):
        """
        Perform depth-first search starting at a node,
        where nodes are visited before their children.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        dq = deque([node])
        while dq:
            current = dq.pop()
            self.visit(current)
            if not self.is_leaf(current):
                for c in reversed(self.children(current)):
                    dq.append(c)
        return self.finalize()

    dfs = dfs_preorder

    def dfs_postorder(self, node):
        """
        Perform depth-first search starting at a node,
        where nodes are visited after their children.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        expanded = set()
        dq = deque([node])
        while dq:
            current = dq[-1]
            if id(current) in expanded:
                dq.pop()
                self.visit(current)
            else:
                for c in reversed(self.children(current)):
                    dq.append(c)
                expanded.add(id(current))
        return self.finalize()

    def dfs_inorder(self, node):
        """
	    Perform depth-first search starting at a root node for
	    a binary tree, where a node is visited after the left 
        child and before the right child.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        expanded = set()
        dq = deque([node])
        while dq:
            current = dq.pop()
            if id(current) in expanded or self.is_leaf(current):
                self.visit(current)
            else:
                first = True
                for c in reversed(self.children(current)):
                    if first:
                        first = False
                    else:
                        dq.append(current)
                    dq.append(c)
                expanded.add(id(current))
        return self.finalize()


class ValueVisitor(object):

    def visit(self, node, values):  #pragma: no cover
        """
	    Visit a node in a tree and compute its value using
        the values of its children.

        This method should be over-written by a user
        that is creating a sub-class.

        Args:
            node: a node in a tree
            values: a list of values of this node's children

        Returns:
            The *value* for this node, which is computed using :attr:`values`
        """
        pass

    def children(self, node):
        """
        Return the children for a node in a tree.
        
        This method has a default implementation, but
        this implementation will likely be redefined
        to reflect the tree data structures used for 
        a specific application.

        Args:
            node: a node in a tree

        Returns:
            A list of children of the specified node.
        """ 
        return node.children
    
    def is_leaf(self, node):
        """
        Return :const:`True` if the node has no children.

        Args:
            node: a node in a tree

        Returns:
            :const:`True` if this node has no children, and :const:`False`
            otherwise
        """
        return len(node.children) == 0

    def finalize(self, ans):    #pragma: no cover
        """
        Return the "final value" of the search.

        The default implementation returns the value of the
        initial node (aka the root node), because
        this visitor pattern computes and returns value for each
        node to enable the computation of this value.

        Args:
            ans: The final value computed by the search method.

        Returns:
            The final value after the search. Defaults to simply
            returning :attr:`ans`.
        """
        return ans

    def visiting_potential_leaf(self, node):
        """ 
        Visit a node and return its value if it is a leaf.

        Args:
            node: a node in a tree

        Returns:
            A tuple: ``(flag, value)``.   If ``flag`` is False,
            then the node is not a leaf and ``value`` is :const:`None`.  
            Otherwise, ``value`` is used to compute the value of the 
            parent node.
        """
        #
        # This default implementation relise on the is_leaf() and visit()
        # methods.  While this provides a simple implementation,
        # in practice a more efficient implementation would 
        # ignore the is_leaf() method and implement this
        # method separately for a specific application.
        #
        if not self.is_leaf(node):
            return False, None
        return True, self.visit(node, None)

    def dfs_postorder_deque(self, node):
        """
        Perform depth-first search starting at a node,
        where nodes are visited after their children.

        This method uses a deque to manage the
        set of nodes that need to be explored.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        flag, value = self.visiting_potential_leaf(node)
        if flag:
            return value
        _values = [[]]
        expanded = set()
        dq = deque([node])
        while dq:
            current = dq[-1]
            if id(current) in expanded:
                dq.pop()
                values = _values.pop()
                _values[-1].append( self.visit(current, values) )
                continue
            flag, value = self.visiting_potential_leaf(current)
            if flag:
                _values[-1].append(value)
                dq.pop()
            else:
                for c in reversed(self.children(current)):
                    dq.append(c)
                expanded.add(id(current))
                _values.append( [] )
        return self.finalize(_values[-1][0])

    def dfs_postorder_stack(self, node):
        """
        Perform depth-first search starting at a node,
        where nodes are visited after their children.

        This method uses a stack to manage the
        set of nodes that need to be explored.

        Args:
            node: a node in a tree

        Returns:
            The value of the :func:`finalize` method.
        """
        flag, value = self.visiting_potential_leaf(node)
        if flag:
            return value
        _stack = [ (node, self.children(node), 0, len(self.children(node)), [])]
        #
        # Iterate until the stack is empty
        #
        # Note: 1 is faster than True for Python 2.x
        #
        while 1:
            #
            # Get the top of the stack
            #   _obj        Current expression object
            #   _argList    The arguments for this expression objet
            #   _idx        The current argument being considered
            #   _len        The number of arguments
            #
            _obj, _argList, _idx, _len, _result = _stack.pop()
            #
            # Iterate through the arguments
            #
            while _idx < _len:
                _sub = _argList[_idx]
                _idx += 1
                flag, value = self.visiting_potential_leaf(_sub)
                if flag:
                    _result.append( value )
                else:
                    #
                    # Push an expression onto the stack
                    #
                    _stack.append( (_obj, _argList, _idx, _len, _result) )
                    _obj                    = _sub
                    _argList                = self.children(_sub)
                    _idx                    = 0
                    _len                    = len(_argList)
                    _result                 = []
            #
            # Process the current node
            #
            ans = self.visit(_obj, _result)
            if _stack:
                #
                # "return" the recursion by putting the return value on the end of the results stack
                #
                _stack[-1][-1].append( ans )
            else:
                return self.finalize(ans)





#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""A plugin that supports global configuration of logging options for pyutilib.component.core."""

import sys
import os.path
import logging
import logging.handlers as handlers

from pyutilib.component.core import Plugin, implements, ExtensionPoint, PluginGlobals
from pyutilib.component.config.env_config import IEnvironmentConfig
from pyutilib.component.config.options import IUpdatedOptionsAction, declare_option


class LoggingConfig(Plugin):
    """A plugin that supports global configuration of logging options."""

    implements(IUpdatedOptionsAction)

    def __init__(self, namespace):
        """Initialize logging information for a specified namespace"""
        self._hdlr = None
        self.namespace = namespace
        self.env_plugins = ExtensionPoint(IEnvironmentConfig)
        if self.namespace == "":
            section = "logging"
            section_re = None
        else:
            section = "logging." + namespace
            section_re = "^logging$"
        #
        declare_option(
            "timestamp",
            section=section,
            section_re=section_re,
            default=False,
            doc="""Add timestamp to logging information.""")
        #
        declare_option(
            "log_dir",
            section=section,
            section_re=section_re,
            default=None,
            doc="""The logging directory.

        The default directory is the application directory plus 'log'.""")
        #
        declare_option(
            "log_type",
            section=section,
            section_re=section_re,
            default='none',
            doc="""Logging facility to use.

        Should be one of (`none`, `file`, `stderr`, `syslog`, `winlog`).""")
        #
        declare_option(
            "log_file",
            section=section,
            section_re=section_re,
            default=namespace + '.log',
            doc="""If `log_type` is `file`, this should be a path to the log-file.""")
        #
        declare_option(
            "log_level",
            section=section,
            section_re=section_re,
            default='WARN',
            doc="""Level of verbosity in log.

        Should be one of (`CRITICAL`, `ERROR`, `WARN`, `INFO`, `DEBUG`).""")
        #
        declare_option(
            "log_format",
            section=section,
            section_re=section_re,
            default=None,
            doc="""Custom logging format.

        If nothing is set, the following will be used:

        $(project)[$(env) $(module)] $(levelname): $(message)

        In addition to regular key names supported by the Python logger library
        library (see http://docs.python.org/lib/node422.html), one could use:
         - $(path)s     the path for the current environment
         - $(basename)s the last path component of the current environment
         - $(app)s      the name of the current application

        Note the usage of `$(...)s` instead of `%(...)s` as the latter form
        would be interpreted by the ConfigParser itself.
        """)

    def reset_after_updates(self):
        """
        Configure the pyutilib.component logging facility.  This will
        implicitly configure all of the environment-specific logging
        objects.
        """
        sys.stdout.flush()
        logger = logging.getLogger('pyutilib.component.core.' + self.namespace)
        if not self._hdlr is None:
            logger.removeHandler(self._hdlr)
        #
        # Set logging level
        #
        level = self.log_level
        level = level.upper()
        if level in ('DEBUG', 'ALL'):
            logger.setLevel(logging.DEBUG)
        elif level == 'INFO':
            logger.setLevel(logging.INFO)
        elif level == 'ERROR':
            logger.setLevel(logging.ERROR)
        elif level == 'CRITICAL':
            logger.setLevel(logging.CRITICAL)
        else:
            logger.setLevel(logging.WARNING)
        #
        # Initialize the current path.  Is there a rule to use for whic
        # environment will be used???  In practice, there is likely to be
        # only one environment.
        #
        if self.log_dir is None:
            path = None
            for plugin in self.env_plugins:
                (flag, count) = plugin.matches(self.namespace)
                tmp = plugin.get_option("path")
                if flag and not tmp is None:
                    path = tmp
                    break
            if path is None:
                path = os.getcwd()
        else:
            path = self.log_dir
        #
        # Setup the logging file
        #
        logtype = self.log_type.lower()
        if self.log_file is None:
            logfile = os.path.join(path, 'log')
        else:
            logfile = self.log_file
            if not os.path.isabs(logfile):
                logfile = os.path.join(path, logfile)
        #
        # Define the format
        #
        format = self.log_format
        if format is None:
            format = '[env=%(env)s where=%(module)s] %(levelname)s - %(message)s'
            if self.timestamp and logtype in ('file', 'stderr'):
                format = '%(asctime)s ' + format
            format = format.replace('$(', '%(') \
                    .replace('%(env)s', PluginGlobals.get_env().name)
        datefmt = ''
        if self.timestamp and self.log_type == 'stderr':
            datefmt = '%X'
        formatter = logging.Formatter(format, datefmt)
        #
        #  Define the handler
        #
        if logtype == 'file':
            hdlr = logging.FileHandler(logfile)
        elif logtype in ('winlog', 'eventlog', 'nteventlog'):
            # Requires win32 extensions
            hdlr = handlers.NTEventLogHandler(logid, logtype='Application')
        elif logtype in ('syslog', 'unix'):
            hdlr = handlers.SysLogHandler('/dev/log')
        elif logtype in ('stderr'):
            hdlr = logging.StreamHandler(sys.stderr)
        else:
            hdlr = handlers.BufferingHandler(0)
            # Note: this _really_ throws away log events, as a `MemoryHandler`
            # would keep _all_ records in case there's no target handler (a bug?)
        self._hdlr = hdlr
        self._logtype = logtype
        self._logfile = logfile
        hdlr.setFormatter(formatter)
        logger.addHandler(hdlr)
        self._log = logger

    def flush(self):
        """Flush logging I/O"""
        self._hdlr.flush()

    def shutdown(self):
        #
        # Q: should this shutdown _all_ logging?
        #
        logging.shutdown()

    def log(self, message):
        self._log.info(message)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['isclose', 'approx_equal', 'as_number', 'isint', 'argmax', 'argmin', 'mean',
           'median', 'factorial', 'perm']

import math
import sys
import six
from six.moves import zip
from six.moves import xrange


def isclose(a, b, rel_tol=1e-9, abs_tol=0.0):
    diff = math.fabs(a-b)
    if diff <= rel_tol*max(math.fabs(a),math.fabs(b)):
        return True
    if diff <= abs_tol:
        return True
    return False


def approx_equal(A, B, abstol, reltol):
    if abstol is None:
        abstol = 1e-8
    if reltol is None:
        reltol = 1e-8
    if math.fabs(A - B) <= abstol:
        return True
    if math.fabs(B) > math.fabs(A):
        relError = math.fabs((A - B) // B)
    else:
        relError = math.fabs((A - B) // A)
    if relError <= reltol:
        return True
    return False


try:
    long

    def as_number(value):
        if type(value) in [int, float, long]:
            return value
        if isinstance(value, six.string_types):
            try:
                tmp = int(value)
                return tmp
            except ValueError:
                pass
            try:
                tmp = long(value)
                return tmp
            except ValueError:
                pass
            try:
                tmp = float(value)
                return tmp
            except ValueError:
                pass
        return value
except:

    def as_number(value):
        if type(value) in [int, float]:
            return value
        if isinstance(value, six.string_types):
            try:
                tmp = int(value)
                return tmp
            except ValueError:
                pass
            try:
                tmp = float(value)
                return tmp
            except ValueError:
                pass
        return value


def isint(arg):
    """
    Returns true if the argument is an integer
    """
    if type(arg) is int:
        return True
    if type(arg) is float:
        tmp = int(arg)
        return (tmp == arg)
    if isinstance(arg, six.string_types):
        try:
            num = float(arg)
            tmp = int(num)
            return (tmp == num)
        except ValueError:
            return False
    return False


def argmax(array):
    """ Return the index to the maximum element of an array """
    return max(zip(array, xrange(len(array))))[1]


def argmin(array):
    """ Return the index to the maximum element of an array """
    return min(zip(array, xrange(len(array))))[1]


def mean(mylist):
    """
    Returns the mean value of a list
    """
    total = 1.0 * sum(mylist)
    length = len(mylist)
    if length == 0.0:
        raise ArithmeticError(
            "Attempting to compute the mean of a zero-length list")
    return (total / length)


if sys.version_info < (3, 0):
    from pyutilib.math.median2 import median
else:
    from pyutilib.math.median3 import median


def factorial(z):
    """
    Computes z!
    """
    if z < 0:
        raise ArithmeticError(
            "Cannot compute the factorial of a negative number")
    if z == 0:
        return 1
    else:
        return z * factorial(z - 1)


def perm(x, y):
    """
    Computes 'x choose y'
    """
    w = 1
    for i in range(y + 1, x + 1):
        w = w * i
    return w / factorial(x - y)



#
# Unit Tests for component/core
#

import re
import sys
import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

from pyutilib.component.core import *
import pyutilib.th as unittest
import pyutilib.misc

try:
    import yaml
    yaml_available = True
except ImportError:
    yaml_available = False

testing_env = PluginEnvironment("testing")
PluginGlobals.add_env(testing_env)


def filter_noncore_interfaces(line):
    return "IConfiguration" in line or \
           "IEnvironmentConfig" in line or \
           "IExternalExecutable" in line or \
           "IFileOption" in line or  \
           "IOption" in line or  \
           "ITempfileManager" in line or  \
           "IWorkflowTask" in line or  \
           "testing" in line or  \
           "IDebug1" in line or  \
           "IDebug2" in line or  \
           "IDebug3" in line or  \
           "IDebug4" in line or  \
           "pyutilib.autotest" in line or  \
           "ITestDriver" in line or  \
           "IFunctorTask" in line or  \
           "ITestParser" in line or  \
           "pyutilib.workflow" in line or  \
           "IUpdatedOptionsAction" in line


class IDebug1(Interface):
    """An example interface"""


class IDebug2(Interface):
    """An example interface"""


class IDebug3(Interface):
    """An example interface"""


class IDebug4(Interface):
    """An example interface that provides default implementations of the api"""

    def __init__(self):
        self.x = 0

    def f1(self):
        self.x = 1

    def f2(self):
        self.x = 2


class Plugin1(Plugin):
    implements(IDebug1, service=True)


class Plugin1a(Plugin):
    implements(IDebug1)


class Plugin2(Plugin):
    implements(IDebug1, service=True)


class Plugin3(Plugin):
    implements(IDebug2, service=True)


class Plugin4(Plugin):
    implements(IDebug1, service=True)
    implements(IDebug2, service=True)


class Plugin5(Plugin):
    implements(IDebug3, service=True)


class Plugin6(Plugin, PluginEnvironment):
    implements(IDebug3, service=True)

    def __init__(self, **kwds):
        Plugin.__init__(self, **kwds)
        PluginEnvironment.__init__(self, "plugin6")


class Plugin7(Plugin):
    implements(IDebug1, service=True)

    def __init__(self, **kwds):
        Plugin.__init__(self, **kwds)
        self._enable = self._id % 2 == 0


class Plugin8(Plugin5):
    implements(IDebug2, service=True)


class Plugin9(Plugin5):
    implements(IDebug3, service=True)


class Service1(PluginEnvironment):

    def is_service_enabled(self, cls):
        return False


class Plugin10(Plugin):
    implements(IDebug1, "tmpenv", service=True)


class Plugin11a(Plugin):
    implements(IDebug4, service=True)

    def __init__(self):
        self.x = 4

    def f1(self):
        self.x = 5


class Plugin11b(Plugin):
    implements(IDebug4, inherit=True, service=True)

    def __init__(self):
        self.x = 0


PluginGlobals.pop_env()


class TestExtensionPoint(unittest.TestCase):

    def setUp(self):
        PluginGlobals.add_env(testing_env)

    def tearDown(self):
        env_ = PluginGlobals.pop_env()
        env_.cleanup(singleton=False)

    def test_interface_decl(self):
        try:

            class IDebug1(Interface):
                pass

            self.fail("expected failure")
        except PluginError:
            pass

    def test_ep_init(self):
        """Test ExtensionPoint construction"""
        ep = ExtensionPoint(IDebug1)
        self.assertEqual(IDebug1, ep.interface)
        try:
            ExtensionPoint()
            self.fail("error expected")
        except PluginError:
            pass

    def test_ep_string(self):
        """Test ExtensionPoint registration"""
        ep = ExtensionPoint(IDebug1)
        self.assertEqual(str(ep), "<ExtensionPoint IDebug1 env=testing>")
        self.assertEqual(len(ep.extensions()), 0)
        s0 = Plugin1()
        self.assertEqual(len(ep.extensions()), 1)
        s1 = Plugin1()
        self.assertEqual(len(ep.extensions()), 2)

    def test_ep_registration(self):
        """Test ExtensionPoint registration"""
        ep = ExtensionPoint(IDebug1)
        self.assertEqual(ep.extensions(), list())
        s1 = Plugin1()
        s1a = Plugin1a()
        s2 = Plugin2()
        s3 = Plugin3()
        s4 = Plugin4()
        s5 = Plugin5()
        self.assertEqual(set(ep.extensions()), set([s1, s2, s4]))
        self.assertEqual(set(ep.extensions()), set([s1, s2, s4]))

    def test_ep_call(self):
        """Test ExtensionPoint __call__"""
        ep = ExtensionPoint(IDebug1)
        s1 = Plugin1(name="p1")
        s2 = Plugin2(name="p2")
        s3 = Plugin4(name="p3")
        s4 = Plugin4(name="p3")
        s5 = Plugin3(name="p4")
        self.assertEqual(
            ep(), sorted(
                set([s1, s2, s3, s4]), key=lambda x: x._id))
        try:
            ep(0)
            self.fail("expected failure")
        except PluginError:
            pass
        self.assertEqual(ep("p1"), [s1])
        self.assertEqual(ep('p3'), sorted(set([s3, s4]), key=lambda x: x._id))

    def test_ep_service(self):
        """Test ExtensionPoint service()"""
        ep = ExtensionPoint(IDebug1)
        s1 = Plugin1()
        s2 = Plugin2()
        s3 = Plugin3()
        s4 = Plugin4()
        try:
            ep.service()
            self.fail("expected failure")
        except PluginError:
            pass

    def test_ep_namespace1(self):
        """Test the semantics of the use of namespaces in interface decl"""
        env = PluginEnvironment("tmpenv")
        s1 = Plugin10()
        s2 = Plugin1()
        namespace_current1 = ExtensionPoint(IDebug1).extensions()
        namespace_current2 = ExtensionPoint(IDebug1, env).extensions()
        self.assertEqual(set(namespace_current1), set((s1, s2)))
        self.assertEqual(set(namespace_current2), set((s1, s2)))


class TestPlugin(unittest.TestCase):

    def setUp(self):
        PluginGlobals.add_env("testing")

    def tearDown(self):
        env_ = PluginGlobals.pop_env()
        env_.cleanup(singleton=False)

    def test_init1(self):
        """Test the behavior of a plugin that is a service manager"""
        s1 = Plugin6()
        self.assertEqual(isinstance(s1, PluginEnvironment), True)
        self.assertEqual(isinstance(s1, Plugin), True)
        self.assertEqual(set(s1.__interfaces__.keys()), set([IDebug3]))

    #def test_init2(self):
    #"""Test that a plugin sets up the registry appropriately"""
    #s1 = Plugin4()
    #s2 = Plugin5()
    #self.assertEqual(s1 in PluginGlobals.interface_registry[IDebug1], True)
    #self.assertEqual(s1 in PluginGlobals.interface_registry[IDebug2], True)
    #self.assertEqual(not s1 in PluginGlobals.extension_points(IDebug3), True)

    def test_init4(self):
        """Verify that base classes are also captured"""
        s1 = Plugin8()
        self.assertEqual(set(s1.__interfaces__.keys()), set([IDebug3, IDebug2]))
        s1 = Plugin9()
        self.assertEqual(set(s1.__interfaces__.keys()), set([IDebug3]))

    def test_init5(self):
        #
        # We don't maintain a registry of environments
        #
        PluginEnvironment("test")
        PluginEnvironment("test")

    def test_repr(self):
        """Test the string representation generated"""
        tmp = PluginGlobals.plugin_counter
        PluginGlobals.plugin_counter = 0
        s1 = Plugin1()
        s2 = Plugin2()
        s3 = Plugin3()
        s4 = Plugin1()
        s5 = Plugin3()
        PluginGlobals.plugin_counter = tmp
        self.assertEqual(str(s1), "<Plugin Plugin1 'Plugin.1'>")
        self.assertEqual(str(s2), "<Plugin Plugin2 'Plugin.2'>")
        self.assertEqual(str(s3), "<Plugin Plugin3 'Plugin.3'>")
        self.assertEqual(str(s4), "<Plugin Plugin1 'Plugin.4'>")
        self.assertEqual(str(s5), "<Plugin Plugin3 'Plugin.5'>")

    def test_enabled(self):
        """Test control of enabled()"""
        ep = ExtensionPoint(IDebug1)
        self.assertEqual(ep.extensions(), list())
        s1 = Plugin1()
        s2 = Plugin2()
        s3 = Plugin3()
        s4 = Plugin4()
        s5 = Plugin5()
        s7a = Plugin7()
        s7b = Plugin7()
        s7c = Plugin7()
        s7d = Plugin7()
        s7e = Plugin7()
        s7f = Plugin7()
        #
        # Only s7b, and s7d will be returned from the exensions() calls
        #
        tmp = set(ep.extensions())
        if s7e in tmp:
            self.assertEqual(
                set(ep.extensions()), set([s1, s2, s4, s7a, s7c, s7e]))
        else:
            self.assertEqual(
                set(ep.extensions()), set([s1, s2, s4, s7b, s7d, s7f]))
        self.assertTrue(PluginGlobals.services() >= set(
            [s1, s2, s3, s4, s5, s7a, s7b, s7c, s7d, s7e, s7f]))

    def test_implements1(self):
        p1 = Plugin11a()
        self.assertEqual(p1.x, 4)
        p1.f1()
        self.assertEqual(p1.x, 5)
        try:
            p1.f2()
            self.fail("Expected AttributeError")
        except AttributeError:
            pass

    def test_implements2(self):
        p1 = Plugin11b()
        self.assertEqual(p1.x, 0)
        p1.f1()
        self.assertEqual(p1.x, 1)
        p1.f2()
        self.assertEqual(p1.x, 2)


class TestMisc(unittest.TestCase):

    def setUp(self):
        PluginGlobals.add_env("testing")

    def tearDown(self):
        env_ = PluginGlobals.remove_env("testing")
        env_.cleanup(singleton=False)

    def test_pprint(self):
        """Test the string representation generated"""

        class Plugin100(SingletonPlugin):
            implements(IDebug3, service=True)

        spx = Plugin100()
        PluginGlobals.add_env("foo")
        s1 = Plugin1()
        s2 = Plugin2()
        s3 = Plugin3()
        tmp_env = PluginGlobals.add_env("bar")
        s4 = Plugin1()
        s5 = Plugin3()
        s6 = Plugin11b()
        sp0 = Plugin100()
        try:
            self.assertFalse(re.match("<Plugin Plugin1", str(s1)) is None)
            self.assertFalse(re.match("<Plugin Plugin2", str(s2)) is None)
            self.assertFalse(re.match("<Plugin Plugin3", str(s3)) is None)
            self.assertFalse(re.match("<Plugin Plugin1", str(s4)) is None)
            self.assertFalse(re.match("<Plugin Plugin3", str(s5)) is None)
            pyutilib.misc.setup_redirect(currdir + "log1.out")
            PluginGlobals.pprint(plugins=False, json=True)
            pyutilib.misc.reset_redirect()
            self.assertMatchesJsonBaseline(currdir + "log1.out",
                                           currdir + "log1.jsn")
            if yaml_available:
                pyutilib.misc.setup_redirect(currdir + "log1.out")
                PluginGlobals.pprint(plugins=False, json=True)
                pyutilib.misc.reset_redirect()
                self.assertMatchesYamlBaseline(currdir + "log1.out",
                                               currdir + "log1.yml")
                #
                pyutilib.misc.setup_redirect(currdir + "log1.out")
                PluginGlobals.pprint(plugins=False)
                pyutilib.misc.reset_redirect()
                self.assertMatchesYamlBaseline(currdir + "log1.out",
                                               currdir + "log1.yml")
        finally:
            PluginGlobals.remove_env("foo", cleanup=True, singleton=False)
            PluginGlobals.remove_env("bar", cleanup=True, singleton=False)


class TestManager(unittest.TestCase):

    def setUp(self):
        PluginGlobals.add_env("testing")

    def tearDown(self):
        env_ = PluginGlobals.pop_env()
        if env_.name == 'testing':
            env_.cleanup(singleton=False)

    def test_init(self):
        """Test the behavior of a plugin that is a service manager"""
        s0 = Plugin1()
        self.assertTrue(PluginGlobals.services() >= set([s0]))
        env = PluginEnvironment("dummy")
        PluginGlobals.add_env(env)
        s1 = Plugin6()
        #self.assertTrue(s1 in PluginGlobals.get_env("testing"))
        #self.assertEqual(s1.services, set([]))
        self.assertTrue(PluginGlobals.services("testing") >= set([s0, s1]))
        PluginGlobals.remove_env("dummy")
        s2 = Plugin6()
        #self.assertTrue(s2 in PluginGlobals.get_env())
        #self.assertEqual(env.services, set([]))
        self.assertTrue(PluginGlobals.services("testing") >= set([s0, s1, s2]))

    def Xtest_get(self):
        env = PluginEnvironment("dummy")
        PluginGlobals.add_env(env)
        s0 = Plugin1()
        self.assertNotEqual(env.active_services(IDebug1), [s0])
        self.assertEqual(
            PluginGlobals.get_env("testing").active_services(IDebug1), [s0])
        try:
            env.active_services(s0)
            self.fail("Expected failure")
        except PluginError:
            pass
        PluginGlobals.pop_env()

    def test_pop1(self):
        """Test that popping the environment doesn't remove the last env"""
        PluginGlobals.pop_env()
        PluginGlobals.pop_env()
        PluginGlobals.pop_env()
        self.assertEqual(len(PluginGlobals.env_stack), 1)

    def test_pop2(self):
        try:
            PluginGlobals.add_env("pca", validate=True)
            self.fail("expected error")
        except PluginError:
            pass
        PluginGlobals.add_env("pca")
        self.assertEqual(PluginGlobals.get_env().name, "pca")
        PluginGlobals.pop_env()

    def test_factory(self):

        class Plugin5_factory(Plugin):
            implements(IDebug3, service=True)

        class Plugin6_factory(Plugin, PluginEnvironment):
            implements(IDebug3, service=True)

            def __init__(self, **kwds):
                Plugin.__init__(self, **kwds)
                PluginEnvironment.__init__(self, "plugin6_factory")

        PluginFactory("Plugin6_factory", name="p6")
        PluginFactory("Plugin5_factory", name="p5")
        PluginFactory("Plugin6_factory")
        try:
            PluginFactory("__foo__")
            self.fail("expected error")
        except PluginError:
            pass
        pyutilib.misc.setup_redirect(currdir + "factory.out")
        PluginGlobals.pprint(plugins=False, json=True)
        pyutilib.misc.reset_redirect()
        self.assertMatchesJsonBaseline(currdir + "factory.out",
                                       currdir + "factory.jsn")
        if yaml_available:
            pyutilib.misc.setup_redirect(currdir + "factory.out")
            PluginGlobals.pprint(plugins=False, json=True)
            pyutilib.misc.reset_redirect()
            self.assertMatchesYamlBaseline(currdir + "factory.out",
                                           currdir + "factory.yml")
            #
            pyutilib.misc.setup_redirect(currdir + "factory.out")
            PluginGlobals.pprint(plugins=False)
            pyutilib.misc.reset_redirect()
            self.assertMatchesYamlBaseline(currdir + "factory.out",
                                           currdir + "factory.yml")


if __name__ == "__main__":
    #import pyutilib.misc
    #print(pyutilib.misc.compare_file(currdir+"log1.out",currdir+"log1.txt", filter=filter_noncore_interfaces))
    unittest.main()



#
# Test simple applications
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
import pyutilib.subprocess


class Test(unittest.TestCase):

    @unittest.skipIf(sys.version_info[:2] < (2, 6), "Skipping tests because configuration output is not guaranteed to be sorted") # yapf: disable
    def test_app1a(self):
        pyutilib.subprocess.run(sys.executable + " " + currdir + os.sep +
                                "app1a.py " + currdir)
        self.assertFileEqualsBaseline(currdir + "config1.out",
                                      currdir + "config1.txt")

    @unittest.skipIf(sys.version_info[:2] < (2, 6), "Skipping tests because configuration output is not guaranteed to be sorted") # yapf: disable
    def test_app1b(self):
        pyutilib.subprocess.run(sys.executable + " " + currdir + os.sep +
                                "app1b.py " + currdir)
        self.assertFileEqualsBaseline(currdir + "config1.out",
                                      currdir + "config1.txt")

    def test_app2(self):
        pyutilib.subprocess.run(sys.executable + " " + currdir + os.sep +
                                "app2.py " + currdir)
        self.assertFileEqualsBaseline(currdir + "summary.out",
                                      currdir + "summary.txt")

    def test_app3(self):
        pyutilib.subprocess.run(sys.executable + " " + currdir + os.sep +
                                "app3.py " + currdir)
        if not os.path.exists(currdir + "app3.log"):
            self.fail("expected log file")
        else:
            os.remove(currdir + "app3.log")

    def test_app4(self):
        pyutilib.subprocess.run(sys.executable + " " + currdir + os.sep +
                                "app4.py " + currdir)
        if not os.path.exists(currdir + "tmp2.ini"):
            self.fail("expected ini file")
        else:
            os.remove(currdir + "tmp2.ini")


if __name__ == "__main__":
    unittest.main()



import pyutilib.misc
import pyutilib.component.core
import os
import sys

currdir = sys.argv[-2] + os.sep

pyutilib.component.core.PluginGlobals.get_env().load_services(
    path=[currdir + "plugins1", currdir + "plugins2"])
pyutilib.misc.setup_redirect(currdir + "load2.out")
if sys.argv[-1] == "json":
    pyutilib.component.core.PluginGlobals.pprint(json=True)
else:
    pyutilib.component.core.PluginGlobals.pprint()
pyutilib.misc.reset_redirect()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['TaskDriver']

import argparse
from pyutilib.misc import Options
from pyutilib.workflow import tasks


class TaskDriver(object):

    def __init__(self, **kwargs):
        if not 'formatter_class' in kwargs:
            kwargs['formatter_class'] = argparse.RawDescriptionHelpFormatter
        self.parser = argparse.ArgumentParser(**kwargs)
        #self.parser.add_argument('--help-commands', dest='help_commands', action='store_true', default=False, help="Print a list of available subcommands")
        self.subparsers = self.parser.add_subparsers(
            help='Sub-commands', dest='subparser_name')
        self.wf = {}

    def register_task(self, wf, name=None, help=''):
        if name is None:
            name = wf
        ans = tasks.TaskFactory(
            wf, parser=self.subparsers.add_parser(
                name, help=help))
        if ans is None:
            raise ValueError(
                "Unknown workflow task '%s'\n    Available tasks: %s" %
                (wf, ' '.join(tasks.TaskFactory().services())))
        ans._parser.set_defaults(subparser_name=name)
        ans.set_arguments()
        self.wf[name] = ans

    def print_help(self):
        self.parser.print_help()  #pragma:nocover

    def parse_args(self, args=None):
        #
        # Call the parser to parse the arguments
        #
        if args is None:
            # sys.argv arguments
            ret = self.parser.parse_args()
        else:
            # args arguments
            ret = self.parser.parse_args(args)
        #
        # Set options, based on the parser return value.
        # The parser options are passed into the workflow.
        #
        opt = Options()
        for key in dir(ret):
            if key[0] != '_' and key != 'subparser_name':
                opt[key] = getattr(ret, key)
        return self.wf[ret.subparser_name](opt)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import logging
import re
import sys
import textwrap

_indention = re.compile('\s*')

class LogHandler(logging.Handler):

    def __init__( self, base='', stream=None,
                  level=logging.NOTSET, verbosity=None ):
        logging.Handler.__init__(self, level=level)

        if verbosity is None:
            verbosity = lambda: True
        if stream is None:
            stream = sys.stdout

        self.verbosity = verbosity
        self.stream = stream
        self.basepath = base
        # Public attributes (because embedded constants in functions are evil)
        self.wrap = 78
        self.initial_indent = ''
        self.subsequent_indent = ' '*4

    def emit(self, record):
        level = record.levelname
        msg = record.getMessage()
        # Most of the messages are either unformatted long lines or
        # triple-quote blocks of text.  In the latter case, if the text
        # starts on the same line as the triple-quote, then it is almost
        # certainly NOT indented with the bulk of the text, which will
        # cause dedent to get confused and not strip any leading
        # whitespace.  This attempts to work around that case:
        #
        #if not ( msg.startswith('\n') or _indention.match(msg).group() ):
        #    # copy the indention for the second line to the first:
        #    lines = msg.splitlines()
        #    if len(lines) > 1:
        #        msg = _indention.match(lines[1]).group() + msg
        #
        # The problem with the above logic is that users may want a
        # simple introductory line followed by an intented line (our
        # tests did this!), and cannot specify it without adding an
        # extra blank line to the output.  In contrast, it is possible
        # for the user to fix the scenario above that motivated this
        # code by just indenting their first line correctly.

        # TBD: dedent does not convert \t to ' '*8. Should we do that?
        msg = textwrap.dedent(msg)

        # As textwrap only works on single paragraphs, we need to break
        # up the incoming message into paragraphs before we pass it to
        # textwrap.
        paragraphs = []
        indent = _indention.match(msg).group()
        par_lines = []
        for line in msg.splitlines():
            leading = _indention.match(line).group()
            content = line.strip()
            if not content:
                paragraphs.append((indent, par_lines))
                par_lines = []
                # Blank lines reset the indentation level
                indent = None
            elif indent == leading:
                # Catch things like bulleted lists
                if len(content) > 1 and par_lines and content[1] == ' ' \
                   and content[0] in '-* ':
                    paragraphs.append((indent, par_lines))
                    par_lines = []
                par_lines.append( content )
            else:
                paragraphs.append((indent, par_lines))
                par_lines = [ content ]
                indent = leading
        # Collect the final paragraph
        if par_lines:
            paragraphs.append((indent, par_lines))

        # Skip any leading/trailing blank lines
        while paragraphs and not paragraphs[-1][1]:
            paragraphs.pop()
        while paragraphs and not paragraphs[0][1]:
            paragraphs.pop(0)

        if self.verbosity():
            #
            # If verbosity is on, the first logged line is the file,
            # line, and function name that called the logger.  The first
            # line of the message is actually the second line of the
            # output (and so is indented/wrapped the same as the rest of
            # the message)
            #
            filename = record.pathname  # file path
            lineno = record.lineno
            try:
                function = record.funcName
            except AttributeError:
                function = '(unknown)'
            if self.basepath and filename.startswith(self.basepath):
                filename = '[base]' + filename[len(self.basepath):]

            self.stream.write(
                '%s: "%s", %d, %s\n' %
                ( level, filename, lineno, function.strip(), ))
        else:
            #
            # If verbosity is off, prepend the log level name to the
            # beginning of the message and format the line without the
            # 'subsequent' indentation of the remainder of the message
            #
            if paragraphs:
                firstPar = ' '.join(paragraphs.pop(0)[1]).strip()
                if level:
                    firstPar = ('%s: %s' % (level, firstPar))
            else:
                firstPar = level
            self.stream.write( '%s\n' % (
                textwrap.fill( firstPar,
                               width=self.wrap,
                               initial_indent=self.initial_indent,
                               subsequent_indent=self.subsequent_indent ), ))
        for indent, par in paragraphs:
            if not indent:
                indent = ''
            # Bulleted lists get indented with a hanging indent
            if par and len(par[0]) > 1 and par[0][0] in '-*':
                hang = ' '*4
            else:
                hang = ''
            self.stream.write( '%s\n' % (
                textwrap.fill(
                    ' '.join(par),
                    width=self.wrap,
                    initial_indent=self.subsequent_indent+indent,
                    subsequent_indent=self.subsequent_indent+indent+hang ), ))


#
# Set up default logging for PyUtilib
#
# __file__ fails if script is called in different ways on Windows
# __file__ fails if someone does os.chdir() before
# sys.argv[0] also fails because it does not always contain the path
from os.path import dirname as _dir, abspath as _abs
import inspect
_pyutilib_base = _dir(_dir(_dir(_abs(inspect.getfile(inspect.currentframe())))))

_logger = logging.getLogger('pyutilib')
_logger.setLevel(logging.WARNING)
_logger.addHandler( LogHandler(
    _pyutilib_base, verbosity=lambda: _logger.isEnabledFor(logging.DEBUG)))



#
# Utilities for working with YAML and JSON data representations.
#
# NOTE: The core utilities treat the YAML and JSON representations as equivalent.  This may
# not be true in general, but it is true for their application within PyUtilib.
#

__all__ = ['yaml_fix', 'json_fix', 'load_yaml', 'load_json', 'extract_subtext',
           'compare_repn', 'compare_strings', 'compare_yaml_files',
           'compare_json_files', 'simple_yaml_parser']

import pprint
import math
import re
import sys
try:
    unicode
except:
    basestring = str
try:  # pragma no cover
    from collections import OrderedDict
except ImportError:  # pragma no cover
    from ordereddict import OrderedDict
try:
    from StringIO import StringIO
except:
    from io import StringIO
try:
    import json
    json_available = True
except:
    json_available = False
try:
    import yaml
    yaml_available = True  #pragma:nocover
except:  #pragma:nocover
    yaml_available = False  #pragma:nocover
from pyutilib.misc.comparison import open_possibly_compressed_file


def yaml_eval(str):
    try:
        val = int(str)
        return val
    except:
        pass
    try:
        val = float(str)
        return val
    except:
        pass
    try:
        val = eval(str)
        return val
    except:
        pass
    return str


def recursive_yaml_parser(stream, _depth=-1):
    depth = -1
    data = None
    for Line in stream:
        line = Line
        flag = True
        while flag:
            flag = False
            if line.strip() == '':
                continue
            if line.strip() == '---':
                continue
            if line.strip() == '...':
                break
            d = 0
            while d < len(line) and line[d] == ' ':
                d += 1
            if line[d] == '#':
                continue
            if depth == -1:
                if d <= _depth:
                    return None, line
                depth = d
            #if d >= depth:
            #depth = d
            if d < depth:
                return data, line
            if d >= depth:
                if line[d] == '-':
                    if data is None:
                        data = []
                    value = line[d + 1:].strip()
                    if len(value) > 0:
                        data.append(yaml_eval(value))
                    else:
                        value, line = recursive_yaml_parser(stream, d)
                        flag = True
                        data.append(value)
                else:
                    if data is None:
                        data = {}
                    tokens = line.split(':')
                    key = tokens[0].strip()
                    value = tokens[1].strip()
                    if len(value) > 0:
                        data[key] = yaml_eval(value)
                    else:
                        value, line = recursive_yaml_parser(stream, d)
                        flag = True
                        data[key] = value
    return data, line


def simple_yaml_parser(stream):
    if isinstance(stream, basestring):
        _stream = open_possibly_compressed_file(stream)
        repn = recursive_yaml_parser(_stream)[0]
        _stream.close()
        return repn
    return recursive_yaml_parser(stream)[0]


def yaml_fix(val):
    if not isinstance(val, basestring):
        return val
    return val.replace(':', '\\x3a')


def json_fix(val):
    return yaml_fix(val)


def extract_subtext(stream, begin_str='', end_str=None, comment='#'):
    if isinstance(stream, basestring):
        _stream = open_possibly_compressed_file(stream)
    else:
        _stream = stream
    if end_str is None:
        end_str == begin_str
    ans = []
    status = len(begin_str) == 0
    for line in _stream:
        tokens = re.split('[\t ]+', line.strip())
        if not status and line.startswith(begin_str):
            status = True
        elif not end_str is None and end_str != '' and line.startswith(end_str):
            break
        elif status:
            if tokens[0] != comment:
                ans.append(line)
    if isinstance(stream, basestring):
        _stream.close()
    return "".join(ans)


def load_yaml(str):
    istream = StringIO(str)
    return yaml.load(istream, Loader=yaml.SafeLoader)


def load_json(str):

    def _to_list(data):
        ans = []
        if sys.version_info < (3, 0):
            for val in data:
                val_type = type(val)
                if val_type is unicode:
                    val = val.encode('utf-8')
                elif val_type is dict:
                    val = _to_dict(val)
                elif val_type is list:
                    val = _to_list(val)
                ans.append(val)
        else:
            for val in data:
                val_type = type(val)
                if val_type is bytes:
                    val = val.encode('utf-8')
                elif val_type is dict:
                    val = _to_dict(val)
                elif val_type is list:
                    val = _to_list(val)
                ans.append(val)
        return ans

    def _to_dict(data):
        ans = {}
        if sys.version_info < (3, 0):
            for key, val in data.iteritems():
                if type(key) is unicode:
                    key = key.encode('utf-8')
                val_type = type(val)
                if val_type is unicode:
                    val = val.encode('utf-8')
                elif val_type is dict:
                    val = _to_dict(val)
                elif val_type is list:
                    val = _to_list(val)
                ans[key] = val
        else:
            for key, val in data.items():
                if type(key) is bytes:
                    key = key.encode('utf-8')
                val_type = type(val)
                if val_type is bytes:
                    val = val.encode('utf-8')
                elif val_type is dict:
                    val = _to_dict(val)
                elif val_type is list:
                    val = _to_list(val)
                ans[key] = val
        return ans

    # Use specialized decoders because JSON returns UNICODE strings,
    # regardless of what string was originally encoded.  We convert 
    # all unicode back to plain str.
    return json.loads(str, object_hook=_to_dict)


def compare_repn(baseline,
                 output,
                 tolerance=0.0,
                 prefix="<root>",
                 exact=True,
                 using_yaml=True):
    if type(baseline) != type(output) and not (
            type(baseline) in [int, float] and type(output) in [int, float]):
        raise IOError(
            "(%s) Structural difference:\nbaseline:\n%s\noutput:\n%s" %
            (prefix, pprint.pformat(baseline), pprint.pformat(output)))
    #
    if type(baseline) is list:
        if not exact and len(baseline) > len(output):
            raise IOError(
                "(%s) Baseline has longer list than output:\nbaseline:\n%s\noutput:\n%s"
                % (prefix, pprint.pformat(baseline), pprint.pformat(output)))
        if exact and len(baseline) != len(output):
            raise IOError(
                "(%s) Baseline list length does not equal output list:\nbaseline:\n%s\noutput:\n%s"
                % (prefix, pprint.pformat(baseline), pprint.pformat(output)))
        j = 0
        i = 0
        msg = ''
        while j < len(baseline) and i < len(output):
            try:
                compare_repn(
                    baseline[j],
                    output[i],
                    tolerance=tolerance,
                    prefix=prefix + "[" + str(i) + "]",
                    exact=exact,
                    using_yaml=using_yaml)
                j += 1
            except Exception:
                msg = sys.exc_info()[1]
                print(msg)
                pass
            i += 1
        if j < len(baseline):
            raise IOError(
                "(%s) Could not find item %d in output list:\nbaseline:\n%s\noutput:\n%s\nERROR: %s"
                % (prefix, j, pprint.pformat(baseline), pprint.pformat(output),
                   msg))
    #
    elif type(baseline) is dict or type(baseline) is OrderedDict:
        if exact and len(baseline.keys()) != len(output.keys()):
            raise IOError(
                "(%s) Baseline and output have different keys:\nbaseline:\n%s\noutput:\n%s"
                % (prefix, pprint.pformat(baseline.keys()),
                   pprint.pformat(output.keys())))
        for key in baseline:
            if not key in output:
                raise IOError(
                    "(%s) Baseline key %s that does not exist in output:baseline:\n%s\noutput:\n%s"
                    % (prefix, key, pprint.pformat(baseline.keys()),
                       pprint.pformat(output.keys())))
            compare_repn(
                baseline[key],
                output[key],
                tolerance=tolerance,
                prefix=prefix + "." + str(key),
                exact=exact,
                using_yaml=using_yaml)
    #
    elif (type(baseline) is float or type(output) is float
         ) and type(baseline) in [int, float] and type(output) in [int, float]:
        if not tolerance is None and math.fabs(baseline - output) > tolerance:
            raise ValueError(
                "(%s) Floating point values differ: baseline=%.17g and output=%.17g (tolerance=%.17g)"
                % (prefix, baseline, output, tolerance))
    elif baseline != output:  #pragma:nocover
        raise ValueError("(%s) Values differ:\nbaseline:\n%s\noutput:\n%s" % (
            prefix, pprint.pformat(baseline), pprint.pformat(output)))


def compare_strings(baseline,
                    output,
                    tolerance=0.0,
                    exact=True,
                    using_yaml=True):
    if using_yaml and not yaml_available:  #pragma:nocover
        raise IOError(
            "Cannot compare YAML strings because YAML is not available")
    if not using_yaml and not json_available:
        raise IOError(
            "Cannot compare JSON strings because JSON is not available")
    if using_yaml:
        baseline_repn = load_yaml(baseline)
        output_repn = load_yaml(output)
    else:
        try:
            baseline_repn = load_json(baseline)
        except Exception:
            print("Problem parsing JSON baseline")
            print(baseline)
            raise
        try:
            output_repn = load_json(output)
        except Exception:
            print("Problem parsing JSON output")
            print(output)
            raise
    compare_repn(
        baseline_repn,
        output_repn,
        tolerance=tolerance,
        exact=exact,
        using_yaml=using_yaml)


def compare_files(baseline_fname,
                  output_fname,
                  tolerance=0.0,
                  baseline_begin='',
                  baseline_end='',
                  output_begin='',
                  output_end=None,
                  exact=True,
                  using_yaml=True):
    INPUT = open_possibly_compressed_file(baseline_fname)
    baseline = extract_subtext(
        INPUT, begin_str=baseline_begin, end_str=baseline_end)
    INPUT.close()
    INPUT = open_possibly_compressed_file(output_fname)
    output = extract_subtext(INPUT, begin_str=output_begin, end_str=output_end)
    INPUT.close()
    compare_strings(
        baseline,
        output,
        tolerance=tolerance,
        exact=exact,
        using_yaml=using_yaml)


def compare_json_files(baseline_fname,
                       output_fname,
                       tolerance=0.0,
                       baseline_begin='',
                       baseline_end='',
                       output_begin='',
                       output_end=None,
                       exact=True):
    return compare_files(
        baseline_fname,
        output_fname,
        tolerance=tolerance,
        baseline_begin=baseline_begin,
        baseline_end=baseline_end,
        output_begin=output_begin,
        output_end=output_end,
        exact=exact,
        using_yaml=False)


def compare_yaml_files(baseline_fname,
                       output_fname,
                       tolerance=0.0,
                       baseline_begin='',
                       baseline_end='',
                       output_begin='',
                       output_end=None,
                       exact=True):
    return compare_files(
        baseline_fname,
        output_fname,
        tolerance=tolerance,
        baseline_begin=baseline_begin,
        baseline_end=baseline_end,
        output_begin=output_begin,
        output_end=output_end,
        exact=exact,
        using_yaml=True)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import sys
from contextlib import contextmanager
from six import StringIO

_old_stdout = []
_old_stderr = []
_local_file = []

try:
    unicode
except:
    basestring = str


def setup_redirect(output):
    """
    Redirect stdout and stderr to a specified output, which
    is either the string name for a file, or a file-like class.
    """
    _old_stdout.append(sys.stdout)
    _old_stderr.append(sys.stderr)
    if isinstance(output, basestring):
        sys.stderr = _Redirecter(output)
        _local_file.append(True)
    else:
        sys.stderr = output
        _local_file.append(False)
    sys.stdout = sys.stderr


def reset_redirect():
    """ Reset redirection to use standard stdout and stderr """
    if len(_old_stdout) > 0:
        if _local_file.pop():
            sys.stdout.close()
        sys.stdout = _old_stdout.pop()
        sys.stderr = _old_stderr.pop()


@contextmanager
def capture_output(output=None):
    """Temporarily redirect stdout into a string buffer."""
    if output is None:
        output = StringIO()
    try:
        setup_redirect(output)
        yield output
    finally:
        reset_redirect()


#
# A class used to manage the redirection of IO.  The sys.stdout and
# sys.stderr values are set to an instance of this class.
#
class _Redirecter:

    def __init__(self, ofile):
        """ Constructor. """
        self.ofile = ofile
        self._out = open(ofile, "w")

    def write(self, s):
        """ Write an item. """
        self._out.write(s)

    def flush(self):
        """ Flush the output. """
        self._out.flush()

    def close(self):
        """ Close the stream. """
        self._out.close()



#
# Unit Tests for pyutilib.autotest
#
#

import os
import sys
from os.path import abspath, dirname, join
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
from pyutilib.misc import setup_redirect, reset_redirect
import pyutilib.autotest
import pyutilib.subprocess
from pyutilib.dev.entry_point import run_entry_point

try:
    import yaml
    yaml_available = True
except ImportError:
    yaml_available = False
try:
    import json
    json_available = True
except ImportError:
    json_available = False


def filter(str):
    return 'currdir' in str or \
           'Ran' in str or \
           'UserWarning' in str


def filter_t1(str):
    return 'options' in str or \
           'Options' in str or \
           'usage' in str or \
           'Usage' in str


class TestYaml(pyutilib.th.TestCase):

    def setUp(self):
        if sys.platform.startswith('java'):
            self.skipTest("Skipping tests because running in Jython")
        self.t1 = os.environ.get('PYUTILIB_AUTOTEST_CATEGORIES', None)
        self.t2 = os.environ.get('PYUTILIB_UNITTEST_CATEGORIES', None)
        if not self.t1 is None:
            del os.environ['PYUTILIB_AUTOTEST_CATEGORIES']
        if not self.t2 is None:
            del os.environ['PYUTILIB_UNITTEST_CATEGORIES']

    def driver(self, *args):
        tmp = ['autotest']
        tmp += list(args)
        pyutilib.autotest.run(tmp, {})

    def tearDown(self):
        if not self.t1 is None:
            os.environ['PYUTILIB_AUTOTEST_CATEGORIES'] = self.t1
        if not self.t2 is None:
            os.environ['PYUTILIB_UNITTEST_CATEGORIES'] = self.t2

    def test1(self):
        # run --help
        setup_redirect(currdir + 'test1.out')
        self.driver('--help')
        reset_redirect()
        self.assertFileEqualsBaseline(
            currdir + 'test1.out', currdir + 'test1.txt', filter=filter_t1)

    def test2(self):
        # run --help-suites example1.yml
        setup_redirect(currdir + 'test2.out')
        self.driver('--help-suites', currdir + 'example1.yml')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test2.out',
                                      currdir + 'test2.txt')

    def test3(self):
        # run --help-categories example1.yml
        setup_redirect(currdir + 'test3.out')
        self.driver('--help-categories', currdir + 'example1.yml')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test3.out',
                                      currdir + 'test3.txt')

    def test4(self):
        # run --help-tests suite1 example1.yml
        setup_redirect(currdir + 'test4.out')
        self.driver('--help-tests', 'suite1', currdir + 'example1.yml')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test4.out',
                                      currdir + 'test4.txt')

    def test5(self):
        # run example1.yml
        run_entry_point('PyUtilib','pyutilib_test_driver',
                        [join(currdir,'example1.yml')],
                        outfile=join(currdir,'test5.out'))
        self.assertFileEqualsBaseline(
            currdir + 'test5.out', currdir + 'test5.txt', filter=filter)

    def test6(self):
        # run --cat x_suite2 --cat x_suite1 example1.yml
        run_entry_point('PyUtilib', 'pyutilib_test_driver',
                        [ '--cat', 'x_suite2', '--cat', 'x_suite1',
                          join(currdir,'example1.yml') ],
                        outfile=join(currdir,'test6.out'))
        self.assertFileEqualsBaseline(
            currdir + 'test6.out', currdir + 'test6.txt', filter=filter)


class TestJson(pyutilib.th.TestCase):

    def setUp(self):
        if sys.platform.startswith('java'):
            self.skipTest("Skipping tests because running in Jython")
        if not json_available:
            self.skipTest("Skipping tests because JSON is not available")
        self.t1 = os.environ.get('PYUTILIB_AUTOTEST_CATEGORIES', None)
        self.t2 = os.environ.get('PYUTILIB_UNITTEST_CATEGORIES', None)
        if not self.t1 is None:
            del os.environ['PYUTILIB_AUTOTEST_CATEGORIES']
        if not self.t2 is None:
            del os.environ['PYUTILIB_UNITTEST_CATEGORIES']

    def driver(self, *args):
        tmp = ['autotest']
        tmp += list(args)
        pyutilib.autotest.run(tmp, {})

    def tearDown(self):
        if not self.t1 is None:
            os.environ['PYUTILIB_AUTOTEST_CATEGORIES'] = self.t1
        if not self.t2 is None:
            os.environ['PYUTILIB_UNITTEST_CATEGORIES'] = self.t2

    def test1(self):
        # run --help
        setup_redirect(currdir + 'test1.out')
        self.driver('--help')
        reset_redirect()
        self.assertFileEqualsBaseline(
            currdir + 'test1.out', currdir + 'test1.txt', filter=filter_t1)

    def test2(self):
        # run --help-suites example1.json
        setup_redirect(currdir + 'test2.out')
        self.driver('--help-suites', currdir + 'example1.json')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test2.out',
                                      currdir + 'test2.txt')

    def test3(self):
        # run --help-categories example1.json
        setup_redirect(currdir + 'test3.out')
        self.driver('--help-categories', currdir + 'example1.json')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test3.out',
                                      currdir + 'test3.txt')

    def test4(self):
        # run --help-tests suite1 example1.json
        setup_redirect(currdir + 'test4.out')
        self.driver('--help-tests', 'suite1', currdir + 'example1.json')
        reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test4.out',
                                      currdir + 'test4.txt')

    def test5(self):
        # run example1.json
        run_entry_point('PyUtilib','pyutilib_test_driver',
                        [join(currdir,'example1.json')],
                        outfile=join(currdir,'test5.out'))
        self.assertFileEqualsBaseline(
            currdir + 'test5.out', currdir + 'test5.txt', filter=filter)

    def test6(self):
        # run --cat x_suite2 --cat x_suite1 example1.json
        run_entry_point('PyUtilib', 'pyutilib_test_driver',
                        [ '--cat', 'x_suite2', '--cat', 'x_suite1',
                          join(currdir,'example1.json') ],
                        outfile=join(currdir,'test6.out'))
        self.assertFileEqualsBaseline(
            currdir + 'test6.out', currdir + 'test6.txt', filter=filter)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

# enum.py
# Part of enum, a package providing enumerated types for Python.
#
# Copyright © 2007 Ben Finney
# This is free software; you may copy, modify and/or distribute this work
# under the terms of the GNU General Public License, version 2 or later
# or, at your option, the terms of the Python license.
#
#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#
# William Hart
# Edits to the PyPI enum package to support __call__ and pickling.
#
""" Robust enumerated type support in Python.

This package provides a module for robust enumerations in Python.

An enumeration object is created with a sequence of string arguments
to the Enum() constructor::

    >>> from enum import Enum
    >>> Colours = Enum('red', 'blue', 'green')
    >>> Weekdays = Enum('mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun')

The return value is an immutable sequence object with a value for each
of the string arguments. Each value is also available as an attribute
named from the corresponding string argument::

    >>> pizza_night = Weekdays[4]
    >>> shirt_colour = Colours.green

The values are constants that can be compared only with values from
the same enumeration; comparison with other values will invoke
Python's fallback comparisons::

    >>> pizza_night == Weekdays.fri
    True
    >>> shirt_colour > Colours.red
    True
    >>> shirt_colour == "green"
    False

Each value from an enumeration exports its sequence index
as an integer, and can be coerced to a simple string matching the
original arguments used to create the enumeration::

    >>> str(pizza_night)
    'fri'
    >>> shirt_colour.index
    2
"""

import six


class EnumException(Exception):
    """ Base class for all exceptions in this module. """

    def __init__(self):
        if self.__class__ is EnumException:
            raise NotImplementedError("%s is an abstract class for subclassing"
                                      % self.__class__)


class EnumEmptyError(AssertionError, EnumException):
    """ Raised when attempting to create an empty enumeration. """

    def __str__(self):
        return "Enumerations cannot be empty"


class EnumBadKeyError(TypeError, EnumException):
    """ Raised when creating an Enum with non-string keys. """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return ("Enumeration keys must be strings or "
                "instances of EnumValue: %s" % (self.key,))


class EnumBadIndexError(AssertionError, EnumException):
    """ Raised when creating an Enum with with an invalid index."""

    def __init__(self, index, reason):
        self.index = index
        self.reason = reason

    def __str__(self):
        return ("Enumeration index (%s) is not valid. Reason: %s" %
                (self.index, self.reason))


class EnumBadTypeError(TypeError, EnumException):
    """ Raised when creating an Enum with a bad type value."""

    def __init__(self, type_, reason):
        self.type_ = type_
        self.reason = reason

    def __str__(self):
        return ("Enumeration type=%r is not valid. Reason: %s" %
                (self.type_, self.reason))


class EnumImmutableError(TypeError, EnumException):
    """ Raised when attempting to modify an Enum. """

    def __init__(self, *args):
        self.args = args

    def __str__(self):
        return "Enumeration does not allow modification"


class EnumValue(object):
    """ A specific value of an enumerated type. """

    def __init__(self, enumtype, index, key):
        """ Set up a new instance. """
        self.__enumtype = enumtype
        self.__index = index
        self.__key = key

    def __get_enumtype(self):
        return self.__enumtype

    enumtype = property(__get_enumtype)

    def __get_key(self):
        return self.__key

    key = property(__get_key)

    def __str__(self):
        return "%s" % (self.key)

    def __get_index(self):
        return self.__index

    index = property(__get_index)

    def __repr__(self):
        return "EnumValue(%s, %s, %s)" % (repr(self.__enumtype),
                                          repr(self.__index),
                                          repr(self.__key),)

    def __hash__(self):
        return hash(self.__index)

    def __lt__(self, other):
        """Less than operator

        (Called in response to 'self < other' or 'other > self'.)
        """
        try:
            return self.index < other.index
        except Exception:
            return NotImplemented

    def __gt__(self, other):
        """Greater than operator

        (Called in response to 'self > other' or 'other < self'.)
        """
        try:
            return self.index > other.index
        except Exception:
            return NotImplemented

    def __le__(self, other):
        """Less than or equal operator

        (Called in response to 'self <= other' or 'other >= self'.)
        """
        try:
            return self.index <= other.index
        except Exception:
            return NotImplemented

    def __ge__(self, other):
        """Greater than or equal operator

        (Called in response to 'self >= other' or 'other <= self'.)
        """
        try:
            return self.index >= other.index
        except Exception:
            return NotImplemented

    def __eq__(self, other):
        """Equal to operator

        (Called in response to 'self == other'.)
        """
        try:
            return self.index == other.index
        except Exception:
            return NotImplemented

    def __ne__(self, other):
        """Equal to operator

        (Called in response to 'self != other'.)
        """
        try:
            return self.index != other.index
        except Exception:
            return NotImplemented


class Enum(object):
    """ Enumerated type. """

    def __init__(self, *keys, **kwargs):
        """ Create an enumeration instance. """

        if not keys:
            raise EnumEmptyError()

        keys = tuple(keys)
        values = [None] * len(keys)

        value_type = kwargs.get('value_type', EnumValue)

        for i, key in enumerate(keys):
            if isinstance(key, six.string_types):
                value = value_type(self, i, key)
            else:
                if not isinstance(key, EnumValue):
                    raise EnumBadKeyError(key)
                value = key
                key = value.key
                if value.index != i:
                    raise EnumBadIndexError(
                        value.index, "Assigned index for argument with key %s, "
                        "does not match location in the initialization list" %
                        (key))
            values[i] = value
            if value.enumtype != values[0].enumtype:
                raise EnumBadTypeError(
                    value.enumtype,
                    "Type assigned to positional argument %s (key=%s) "
                    "does not match the type assigned to the first "
                    "positional argument (key=%s): %r" %
                    (i, key, values[0].key, values[0].enumtype))
            try:
                super(Enum, self).__setattr__(key, value)
            except TypeError:
                raise EnumBadKeyError(key)

        super(Enum, self).__setattr__('_keys', keys)
        super(Enum, self).__setattr__('_values', values)

    def __call__(self, index):
        #
        # If the index is an integer, get the item
        # with this index
        #
        if isinstance(index, int):
            return self[index]
        #
        # If the index is not a string, then try coercing it
        #
        if not isinstance(index, six.string_types):
            tmp = str(index)
        else:
            tmp = index
        return getattr(self, tmp)

    def __setattr__(self, name, value):
        raise EnumImmutableError(name)

    def __delattr__(self, name):
        raise EnumImmutableError(name)

    def __len__(self):
        return len(self._values)

    def __getitem__(self, index):
        return self._values[index]

    def __setitem__(self, index, value):
        raise EnumImmutableError(index)

    def __delitem__(self, index):
        raise EnumImmutableError(index)

    def __iter__(self):
        return iter(self._values)

    def __contains__(self, value):
        is_member = False
        if isinstance(value, six.string_types):
            is_member = (value in self._keys)
        else:
            try:
                is_member = (value in self._values)
            except Exception:
                is_member = False
        return is_member
